#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-agnostic type hint reduction** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._check.convert._reduce.redmain` submodule.
'''

# ....................{ TODO                               }....................
#FIXME: We've currently split some PEP-specific hint_reduce() unit tests into
#PEP-specific submodules of this subpackage (e.g., "pep.test_redpep695"). The
#the test_reduce() unit test defined below is dramatically superior, however.
#Why? Internal dataclass machinery uniquely implemented in this test automates
#away associated unpleasantness. Thus, for maintainability, *ALL* of those
#submodules should be:
#* Integrated into the sole test_reduce_hint() unit test defined below. Yes,
#  this test will begin to become uncomfortably long. Who cares? Well,
#  technically, *WE* care -- but not that much. Long tests can later be
#  subdivided. The critical point is that test_reduce() machinery is superior.
#* Subsequently removed.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytmark import ignore_warnings

# ....................{ TESTS ~ reducers                   }....................
#FIXME: This unit test is getting a bit long in the tooth. Consider splitting
#into smaller PEP-specific unit tests using the same pseudo-fixture approach as
#implemented by the "beartype_test.a00_unit.data.hint.pep.data_pepfixture" submodule.
def test_reduce_hint() -> None:
    '''
    Test the private
    :func:`beartype._check.convert._reduce.redmain.reduce_hint` reducer.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer test-specific imports.
    from beartype.typing import (
        Annotated,
    )
    from beartype.vale import IsEqual
    from beartype._cave._cavefast import NoneType
    from beartype._check.convert._reduce.redmain import reduce_hint
    from beartype._conf.confcommon import BEARTYPE_CONF_DEFAULT
    from beartype._conf.confmain import BeartypeConf
    from beartype._data.cls.datacls import TYPES_PEP484_GENERIC_IO
    from beartype._data.typing.datatyping import (
        Pep484TowerComplex,
        Pep484TowerFloat,
        TypeException,
    )
    from beartype._data.typing.datatypingport import Hint
    from beartype._data.hint.sign.datahintsigns import HintSignAnnotated
    from beartype._data.kind.datakindiota import SENTINEL
    from beartype._util.hint.pep.proposal.pep484604 import (
        make_hint_pep484604_union)
    from beartype._util.hint.pep.proposal.pep593 import is_hint_pep593
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11
    from beartype_test.a00_unit.data.pep.data_pep484 import (
        T_str_or_bytes)
    from beartype_test._util.module.pytmodtest import is_package_numpy
    from dataclasses import InitVar
    from pytest import raises

    # Intentionally import PEP 484-specific type hint factories.
    from typing import Protocol

    # ..................{ CLASSES                            }..................
    class HintReductionValid(object):
        '''
        **Valid type hint reduction case** (i.e., dataclass encapsulating the
        valid use case of reducing one type hint into another).

        Attributes
        ----------
        conf : BeartypeConf
            Input beartype configuration configuring this reduction.
        hint_unreduced : Hint
            Input hint to be reduced.
        hint_reduced : Hint
            Output hint expected to be returned from the :func:`.reduce_hint`
            reducer when passed these inputs.
        '''

        # ..................{ INITIALIZERS                   }..................
        def __init__(
            self,

            # Mandatory parameters.
            hint_unreduced: Hint,

            # Optional parameters.
            conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
            hint_reduced: Hint = SENTINEL,
        ) -> None:
            '''
            Initialize this valid type hint reduction case.

            Attributes
            ----------
            hint_unreduced : Hint
                Input hint to be reduced.
            conf : BeartypeConf, default: BEARTYPE_CONF_DEFAULT
                Input beartype configuration configuring this reduction.
                Defaults to the default beartype configuration.
            hint_reduced : Hint, default: SENTINEL
                Output hint expected to be returned from the
                :func:`.reduce_hint` reducer when passed these inputs. Defaults
                to the sentinel placeholder, in which case this output hint
                actually defaults to this input hint. This default trivializes
                testing for **irreducible hints** (i.e., hints *not* reduced by
                the :func:`.reduce_hint` reducer).
            '''

            # If unpassed, default this output hint to this input hint.
            if hint_reduced is SENTINEL:
                hint_reduced = hint_unreduced

            # Classify all passed parameters.
            self.hint_unreduced = hint_unreduced
            self.hint_reduced = hint_reduced
            self.conf = conf


    class HintReductionInvalid(object):
        '''
        **Invalid type hint reduction case** (i.e., dataclass encapsulating the
        invalid use case of reducing one type hint, which then raises an
        exception due to being invalid).

        Attributes
        ----------
        conf : BeartypeConf
            Input beartype configuration configuring this reduction.
        hint_unreduced : Hint
            Input hint to be reduced.
        exception_type : Type[Exception]
            Output type of exception expected to be raised by the
            :func:`.reduce_hint` reducer when passed these inputs.
        '''

        # ..................{ INITIALIZERS                   }..................
        def __init__(
            self,

            # Mandatory parameters.
            hint_unreduced: Hint,
            exception_type: TypeException,

            # Optional parameters.
            conf: BeartypeConf = BEARTYPE_CONF_DEFAULT,
        ) -> None:
            '''
            Initialize this invalid type hint reduction case.

            Attributes
            ----------
            hint_unreduced : Hint
                Input hint to be reduced.
            exception_type : Type[Exception]
                Output type of exception expected to be raised by the
                :func:`.reduce_hint` reducer when passed these inputs.
            conf : BeartypeConf, default: BEARTYPE_CONF_DEFAULT
                Input beartype configuration configuring this reduction.
                Defaults to the default beartype configuration.
            '''

            # Classify all passed parameters.
            self.hint_unreduced = hint_unreduced
            self.exception_type = exception_type
            self.conf = conf

    # ..................{ LOCALS                             }..................
    # List of all valid reduction cases to be tested.
    hint_reductions_valid = [
        # ..................{ PEP 484                        }..................
        # An isinstanceable type is preserved as is without reduction.
        HintReductionValid(int),

        # The builtin "None" singleton reduces to its type (i.e., "type(None)").
        HintReductionValid(hint_unreduced=None, hint_reduced=NoneType),

        # ..................{ PEP 484 ~ tower                }..................
        # The builtin "float" and "complex" types reduce to their corresponding
        # numeric towers when configured to do so.
        HintReductionValid(
            hint_unreduced=float,
            hint_reduced=Pep484TowerFloat,
            conf=BeartypeConf(is_pep484_tower=True),
        ),
        HintReductionValid(
            hint_unreduced=complex,
            hint_reduced=Pep484TowerComplex,
            conf=BeartypeConf(is_pep484_tower=True),
        ),

        # The builtin "float" and "complex" types are preserved as is without
        # being reduced when configured to do so.
        HintReductionValid(
            hint_unreduced=float,
            hint_reduced=float,
            conf=BeartypeConf(is_pep484_tower=False),
        ),
        HintReductionValid(
            hint_unreduced=complex,
            hint_reduced=complex,
            conf=BeartypeConf(is_pep484_tower=False),
        ),

        # ..................{ PEP 484 ~ typevar              }..................
        # A PEP 484-compliant constrained type variable reduces to the PEP 484-
        # or 604-compliant union of those constraints.
        HintReductionValid(
            hint_unreduced=T_str_or_bytes,
            hint_reduced=make_hint_pep484604_union((str, bytes,)),
        ),

        # ..................{ PEP 557                        }..................
        # A PEP 557-compliant "InitVar" is reduced to its child hint.
        HintReductionValid(hint_unreduced=InitVar[str], hint_reduced=str),

        # ..................{ PEP 593                        }..................
        # A PEP 593-compliant beartype-agnostic metahint is reduced to the
        # lower-level hint it annotates.
        HintReductionValid(hint_unreduced=Annotated[int, 42], hint_reduced=int),

        # A PEP 593-compliant beartype-specific metahint is preserved as is.
        HintReductionValid(Annotated[str, IsEqual['In their noonday dreams.']]),
    ]

    # List of all invalid reduction cases to be tested.
    hint_reductions_invalid = []

    # ..................{ PEP 646                            }..................
    # If the active Python interpreter targets Python >= 3.11 and thus supports
    # PEP 646...
    if IS_PYTHON_AT_LEAST_3_11:
        # ....................{ IMPORTS                    }....................
        # Defer PEP-specific imports.
        from beartype.roar import BeartypeDecorHintPep646Exception
        from beartype._util.hint.pep.proposal.pep646692 import (
            make_hint_pep646_tuple_unpacked_prefix)
        from beartype_test.a00_unit.data.pep.data_pep646 import (
            Ts_unpacked,
            Us_unpacked,
        )

        # ....................{ CLASSES                    }....................
        # class GloomBird(TypedDict):
        #     '''
        #     Arbitrary :pep:`589`-compliant typed dictionary.
        #     '''
        #
        #     pass

        # ....................{ LOCALS                     }....................
        # Extend this list with PEP 646-compliant valid reduction cases.
        hint_reductions_valid.extend((
            # A PEP 646-compliant tuple hint subscripted by *ONLY* a single PEP
            # 646-compliant unpacked type variable tuple reduces to the
            # semantically equivalent builtin "tuple" type.
            HintReductionValid(
                hint_unreduced=tuple[Ts_unpacked], hint_reduced=tuple),

            # A PEP 646-compliant tuple hint subscripted by *ONLY* a single PEP
            # 646-compliant unpacked child variable-length tuple hint reduces to
            # the semantically equivalent PEP 585-compliant variable-length
            # tuple hint subscripted by the same child hints as that unpacked
            # child hint.
            HintReductionValid(
                hint_unreduced=(
                    tuple[make_hint_pep646_tuple_unpacked_prefix((str, ...))]),
                hint_reduced=tuple[str, ...],
            ),

            # A PEP 646-compliant tuple hint subscripted by a PEP 646-compliant
            # unpacked child fixed-length tuple hint reduces to the semantically
            # equivalent PEP 585-compliant fixed-length tuple hint.
            HintReductionValid(
                hint_unreduced=tuple[
                    int,
                    make_hint_pep646_tuple_unpacked_prefix((str, bool)),
                    float,
                ],
                hint_reduced=tuple[int, str, bool, float],
            ),
        ))

        # Extend this list with PEP 646-compliant invalid reduction cases.
        hint_reductions_invalid.extend((
            #FIXME: *UNCOMMENT*, please. This is totally invalid but no longer
            #covered by the current implementation of this reducer. Whatevahs!
            # # A PEP 646-compliant tuple hint subscripted by a PEP 692-compliant
            # # unpacked type dictionary is invalid.
            # (
            #     tuple[Unpack[GloomBird]],
            #     BeartypeDecorHintPep646Exception,
            # ),

            # A PEP 646-compliant tuple hint subscripted by two PEP
            # 646-compliant unpacked child fixed-length tuple hint separated by
            # other unrelated child hints is invalid.
            HintReductionInvalid(
                hint_unreduced=tuple[
                    str,
                    make_hint_pep646_tuple_unpacked_prefix((int, float)),
                    bool,
                    make_hint_pep646_tuple_unpacked_prefix((complex, list)),
                    bytes,
                ],
                exception_type=BeartypeDecorHintPep646Exception,
            ),

            # A PEP 646-compliant tuple hint subscripted by two PEP 646-compliant
            # unpacked type variable tuples separated by other unrelated child hints
            # is invalid.
            HintReductionInvalid(
                hint_unreduced=(
                    tuple[str, Ts_unpacked, bool, Us_unpacked, bytes]),
                exception_type=BeartypeDecorHintPep646Exception,
            ),

            # A PEP 646-compliant tuple hint subscripted by one PEP
            # 646-compliant unpacked child variable-length tuple hint *AND* one
            # PEP 646-compliant unpacked child type variable tuple separated by
            # other unrelated child hints is invalid.
            #
            # Order is probably insignificant -- but could be. Ergo, we test
            # both orders for fuller coverage.
            HintReductionInvalid(
                hint_unreduced=tuple[
                    str,
                    make_hint_pep646_tuple_unpacked_prefix((complex, float)),
                    bool,
                    Ts_unpacked,
                    bytes,
                ],
                exception_type=BeartypeDecorHintPep646Exception,
            ),
            HintReductionInvalid(
                hint_unreduced=tuple[
                    str,
                    Ts_unpacked,
                    bool,
                    make_hint_pep646_tuple_unpacked_prefix((float, int)),
                    bytes,
                ],
                exception_type=BeartypeDecorHintPep646Exception,
            ),
        ))
    # Else, the active Python interpreter targets Python < 3.11 and thus fails
    # to support PEP 646.

    # ....................{ PASS                           }....................
    # For each valid hint reduction case...
    for hint_reduction_valid in hint_reductions_valid:
        # Sanified metadata encapsulating the reduction of this input hint.
        hint_reduced_sane = reduce_hint(
            hint=hint_reduction_valid.hint_unreduced,
            conf=hint_reduction_valid.conf,
        )

        # Assert that this reduction produced the expected output hint.
        assert hint_reduced_sane.hint == hint_reduction_valid.hint_reduced

    # ....................{ FAIL                           }....................
    # For each invalid hint reduction case...
    for hint_reduction_invalid in hint_reductions_invalid:
        # Assert that this reducer raises the expected type of exception when
        # passed this input hint.
        with raises(hint_reduction_invalid.exception_type):
            reduce_hint(
                hint=hint_reduction_invalid.hint_unreduced,
                conf=hint_reduction_invalid.conf,
            )

    # ..................{ PEP 544                            }..................
    # For each PEP 484-compliant "typing" IO generic superclass...
    for pep484_generic_io in TYPES_PEP484_GENERIC_IO:
        # Metadata encapsulating the equivalent PEP 544-compliant protocol
        # reduced from this generic.
        pep544_protocol_io_sane = reduce_hint(pep484_generic_io)

        # This protocol.
        pep544_protocol_io = pep544_protocol_io_sane.hint

        # Assert this protocol is either...
        assert (
            # A PEP 593-compliant type metahint generalizing a protocol *OR*...
            is_hint_pep593(pep544_protocol_io) or
            # A PEP 544-compliant protocol.
            issubclass(pep544_protocol_io, Protocol)
        )

    # ..................{ API ~ numpy                        }..................
    # If a recent version of NumPy is importable...
    if is_package_numpy():
        # Defer API-specific imports.
        from beartype.roar import BeartypeDecorHintNonpepNumpyException
        from numpy import float64
        from numpy.typing import NDArray

        # Metadata encapsulating the reduction of a NumPy array type hint to a
        # beartype validator.
        ndarray_reduced_sane = reduce_hint(NDArray[float64])

        # This beartype validator.
        ndarray_reduced = ndarray_reduced_sane.hint

        # Assert this validator is a "typing{_extensions}.Annotated" hint.
        assert get_hint_pep_sign(ndarray_reduced) is HintSignAnnotated

        # Assert that reducing a "numpy.typing.NDArray" type hint
        # erroneously subscripted by an object *NOT* reducible to a data
        # type raises the expected exception.
        with raises(BeartypeDecorHintNonpepNumpyException):
            reduce_hint(NDArray['From_my_wings_are_shaken_the_dews_that_waken'])

# ....................{ TESTS ~ raiser                     }....................
# Prevent pytest from capturing and displaying all expected non-fatal
# beartype-specific warnings emitted by this test. Urgh!
@ignore_warnings(DeprecationWarning)
def test_reduce_hint_ignorable(hints_pep_meta, hints_ignorable) -> None:
    '''
    Test the private
    :func:`beartype._check.convert._reduce.redmain.reduce_hint` reducer with
    respect to ignorable hints.

    Parameters
    ----------
    hints_pep_meta : tuple[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        Tuple of type hint metadata describing sample type hints exercising edge
        cases in the :mod:`beartype` codebase.
    hints_ignorable : frozenset
        Frozen set of ignorable PEP-agnostic type hints.
    '''

    # Defer test-specific imports.
    from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
    from beartype._check.convert._reduce.redmain import reduce_hint
    from beartype_test.a00_unit.data.hint.data_hint import (
        HINTS_NONPEP_UNIGNORABLE)

    # Assert this tester accepts ignorable type hints.
    for hint_ignorable in hints_ignorable:
        assert reduce_hint(hint_ignorable) is HINT_SANE_IGNORABLE

    # Assert this tester rejects unignorable PEP-noncompliant type hints.
    for hint_unignorable in HINTS_NONPEP_UNIGNORABLE:
        assert reduce_hint(hint_unignorable) is not HINT_SANE_IGNORABLE

    # Assert this tester:
    # * Accepts unignorable PEP-compliant type hints.
    # * Rejects ignorable PEP-compliant type hints.
    for hint_pep_meta in hints_pep_meta:
        # True only if this hint reduces to the ignorable "Any" singleton.
        is_hint_ignorable = reduce_hint(hint_pep_meta.hint) is HINT_SANE_IGNORABLE

        # Assert this hint is either ignorable or unignorable as expected.
        assert hint_pep_meta.is_ignorable == is_hint_ignorable



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **class setters** (i.e., low-level callables modifying various
properties of arbitrary classes).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from collections.abc import Callable

# ....................{ SETTERS                            }....................
#FIXME: Unit test us up.
def set_type_attr(
    # Mandatory parameters.
    cls: type,
    attr_name: str,
    attr_value: object,

    # Optional parameters.
    setattr_func: Callable[[object, str, object], None] = setattr,
) -> None:
    '''
    Dynamically set the **class variable** (i.e., attribute of the passed class)
    with the passed name to the passed value.

    Parameters
    ----------
    cls : type
        Class to set this attribute on.
    attr_name : str
        Name of the class attribute to be set.
    attr_value : object
        Value to set this class attribute to.
    setattr_func : Callable[[object, str, object], None], default: setattr
        :func:`setattr`-like callable to be called to set this attribute on this
        class. Defaults to the C-based :func:`setattr` builtin, which internally
        defers to the ``__setattr__()`` dunder method guaranteed to be defined
        by all classes (due to the existence of the :class:`type.__setattr__`
        dunder method). Though this default suffices for most use cases, a small
        subset of outlier edge cases require that this setter call a non-default
        :func:`setattr`-like callable.

        :pep:`557-compliant frozen dataclasses are the canonical example. They
        guarantee immutability by overriding the ``__setattr__()`` dunder method
        (implicitly called by the :func:`setattr` builtin) to unconditionally
        raise an exception. While understandable, this behaviour prevents the
        :func:`beartype.beartype` decorator from monkey-patching type-checking
        into frozen dataclasses by replacing their default ``__setattr__()``
        with a :func:`beartype.beartype`-specific alternative performing
        type-checking; attempting to do so ironically invokes that same
        ``__setattr__()`, which then raises an exception. This behaviour can be
        circumvented by passing the :func:`type.__setattr__` dunder method as
        this parameter, which then applies the desired monkey-patch *without*
        raising an exception. In short, stupid kludges is always the answer.

    Caveats
    -------
    **This function is unavoidably slow.** Class attributes are *only* settable
    by calling the tragically slow :func:`setattr` builtin. Attempting to
    directly set an attribute on the class dictionary raises an exception: e.g.,

    .. code-block:: pycon

       >>> class OhGodHelpUs(object): pass
       >>> OhGodHelpUs.__dict__['even_god_cannot_help'] = 2
       TypeError: 'mappingproxy' object does not support item assignment

    Why? Because class dictionaries are actually low-level :class:`mappingproxy`
    objects that intentionally override the ``__setattr__()`` dunder method to
    unconditionally raise an exception. Why? Because that constraint enables the
    :meth:`type.__setattr__` dunder method to enforce critical efficiency
    constraints on class attributes -- including that class attribute keys are
    *not* only strings but also valid Python identifiers:

    See also this `relevant StackOverflow answer by Python luminary Raymond
    Hettinger <answer_>`__.

    .. _answer:
       https://stackoverflow.com/a/32720603/2809027
    '''
    assert isinstance(cls, type), f'{repr(cls)} not type.'
    assert callable(setattr_func), f'{repr(setattr_func)} uncallable.'

    # Attempt to set the class attribute with this name to this value.
    try:
        setattr_func(cls, attr_name, attr_value)
    # If doing so raises a builtin "TypeError"...
    except TypeError as exception:
        # Message raised with this "TypeError".
        exception_message = str(exception)

        # If this message satisfies a well-known pattern unique to the current
        # Python version, then this exception signifies this attribute to be
        # inherited from an immutable builtin type (e.g., "str") subclassed by
        # this user-defined subclass. In this case, silently skip past this
        # uncheckable attribute to the next. Specifically, since the active
        # Python interpreter targets Python >= 3.10, match a message of the
        # form:
        #     cannot set '{attr_name}' attribute of immutable type '{cls_name}'
        if (
            exception_message.startswith("cannot set '") and
            "' attribute of immutable type " in exception_message
        ):
            return
        # Else, this message does *NOT* satisfy that pattern.

        # Preserve this exception by re-raising this exception.
        raise



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator call metadata dataclass** (i.e., class aggregating *all*
metadata for the callable currently being decorated by the
:func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorWrappeeException
from beartype.typing import (
    TYPE_CHECKING,
    Callable,
    FrozenSet,
    Optional,
)
from beartype._cave._cavefast import CallableCodeObjectType
from beartype._cave._cavemap import NoneTypeOr
from beartype._check.forward.fwdscope import BeartypeForwardScope
from beartype._conf.confmain import BeartypeConf
from beartype._data.code.pep.datacodepep525 import CODE_PEP525_RETURN_PREFIX
from beartype._data.typing.datatyping import (
    LexicalScope,
    Pep649HintableAnnotations,
    TypeStack,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.cache.pool.utilcachepoolinstance import (
    acquire_instance,
    release_instance,
)
from beartype._util.func.utilfunccodeobj import (
    get_func_codeobj,
    get_func_codeobj_or_none,
)
from beartype._util.func.utilfunctest import (
    is_func_coro,
    is_func_nested,
    is_func_sync_generator,
    is_func_async_generator,
)
from beartype._util.func.utilfuncwrap import unwrap_func_all_isomorphic
from beartype._util.hint.pep.proposal.pep649 import (
    get_pep649_hintable_annotations,
    set_pep649_hintable_annotations,
)
from beartype._util.text.utiltextprefix import prefix_callable_pith

# ....................{ CLASSES                            }....................
class BeartypeDecorMeta(object):
    '''
    **Beartype decorator call metadata** (i.e., object encapsulating *all*
    metadata for the callable currently being decorated by the
    :func:`beartype.beartype` decorator).

    Design
    ------
    This the *only* object instantiated by that decorator for that callable,
    substantially reducing both space and time costs. That decorator then
    passes this object to most lower-level functions, which then:

    #. Access read-only instance variables of this object as input.
    #. Modify writable instance variables of this object as output. In
       particular, these lower-level functions typically accumulate pure-Python
       code comprising the generated wrapper function type-checking the
       decorated callable by setting various instance variables of this object.

    Caveats
    -------
    **This object cannot be used to communicate state between low-level
    memoized callables** (e.g.,
    :func:`beartype._check.code.codemain.make_func_pith_code`) **and
    high-level unmemoized callables** (e.g.,
    :func:`beartype._decor._nontype._wrap.wrapmain.generate_code`). Instead,
    low-level memoized callables *must* return that state as additional return
    values up the call stack to those high-level unmemoized callables. By
    definition, memoized callables are *not* recalled on subsequent calls passed
    the same parameters. Since only the first call to those callables passed
    those parameters would set the appropriate state on this object intended to
    be communicated to unmemoized callables, *all* subsequent calls would subtly
    fail with difficult-to-diagnose issues. See also `<issue #5_>`__, which
    exhibited this very complaint.

    .. _issue #5:
       https://github.com/beartype/beartype/issues/5

    Attributes
    ----------
    cls_stack : TypeStack
        **Type stack** (i.e., either tuple of zero or more arbitrary types *or*
        :data:`None`). See also the parameter of the same name accepted by the
        :func:`beartype._decor.decorcore.beartype_object` function for details.
    conf : BeartypeConf
        **Beartype configuration** (i.e., self-caching dataclass encapsulating
        all flags, options, settings, and other metadata configuring the
        current decoration of the decorated callable).
    func_annotations : dict[str, Hint]
        **Type hint dictionary** mapping from the name of each annotated
        parameter and return accepted by the decorated callable to the type hint
        annotating that parameter or return.

        Note that this dictionary is *not* directly mutable. When the active
        Python interpreter targets Python >= 3.14, :pep:`649`-compliant type
        hint dictionaries are *not* directly mutable. Attempting to do so will
        superficially appear to succeed but ultimately reduce to a silent noop.
        This particular dictionary is only indirectly mutable by calling the
        high-level :meth:`set_func_pith_hint` setter method, which mutates the
        type hint annotating the name of the passed parameter or return in a
        portable manner consistent with both :pep:`649` and Python >= 3.14.
        Although this constraint *could* be enforced by encapsulating this
        dictionary in a :class:`beartype.FrozenDict` instance, doing so would
        simply reduce space and time efficiency for little to no actual gain.
        Simply call :meth:`set_func_pith_hint` instead.
    func_annotations_get : Callable[[str, object], object]
        :meth:`dict.get` method bound to the :attr:`func_annotations`
        dictionary, localized as a negligible microoptimization. Blame Guido.
    _is_func_annotations_dirty : bool
        :data:`True` only if the type hint dictionary is **dirty** (i.e.,
        modified from the original type hint dictionary annotating the decorated
        callable by a prior call to the :meth:`set_func_pith_hint` setter).
    func_wrappee : Callable
        Possibly wrapping **decorated callable** (i.e., high-level callable
        currently being decorated by the :func:`beartype.beartype` decorator).
        Note the lower-level :attr:`func_wrappee_wrappee` callable should
        *usually* be accessed instead; although higher-level, this callable may
        only be a wrapper function and hence yield inaccurate or even erroneous
        metadata (especially the code object) for the callable being wrapped.
    func_wrappee_is_nested : bool
        Either:

        * If this wrappee callable is **nested** (i.e., declared in the body of
          another pure-Python callable or class), :data:`True`.
        * If this wrappee callable is **global** (i.e., declared at module scope
          in its submodule), :data:`False`.
    func_wrappee_scope_forward : Optional[BeartypeForwardScope]
        Either:

        * If this wrappee callable is annotated by at least one **stringified
          type hint** (i.e., declared as a :pep:`484`- or :pep:`563`-compliant
          forward reference referring to an actual type hint that has yet to be
          declared in the local and global scopes declaring this callable) that
          :mod:`beartype` has already resolved to its referent, this wrappee
          callable's **forward scope** (i.e., dictionary mapping from the name
          to value of each locally and globally accessible attribute in the
          local and global scope of this wrappee callable as well as deferring
          the resolution of each currently undeclared attribute in that scope by
          replacing that attribute with a forward reference proxy resolved only
          when that attribute is passed as the second parameter to an
          :func:`isinstance`-based runtime type-check).
        * Else, :data:`None`.

        Note that:

        * The reconstruction of this scope is computationally expensive and thus
          deferred until needed to resolve the first stringified type hint
          annotating this wrappee callable.
        * All callables have local scopes *except* global functions, whose local
          scopes are by definition the empty dictionary.
    func_wrappee_scope_nested_names : Optional[frozenset[str]]
        Either:

        * If this wrappee callable is annotated by at least one stringified type
          hint that :mod:`beartype` has already resolved to its referent,
          either:

          * If this wrappee callable is **nested** (i.e., declared in the body
            of another pure-Python callable or class), the non-empty frozen set
            of the unqualified names of all parent callables lexically
            containing this nested wrappee callable (including this nested
            wrappee callable itself).
          * Else, this wrappee callable is declared at global scope in its
            submodule. In this case, the empty frozen set.

        * Else, :data:`None`.
    func_wrappee_wrappee : Callable
        Possibly unwrapped **decorated callable wrappee** (i.e., low-level
        callable wrapped by the high-level :attr:`func_wrappee` callable
        currently being decorated by the :func:`beartype.beartype` decorator).
        If the higher-level :attr:`func_wrappee` callable does *not* actually
        wrap another callable, this callable is identical to that callable.
    func_wrappee_wrappee_codeobj : CallableCodeObjectType
        Possibly unwrapped **decorated callable wrappee code object** (i.e.,
        code object underlying the low-level :attr:`func_wrappee_wrappee`
        callable wrapped by the high-level :attr:`func_wrappee` callable
        currently being decorated by the :func:`beartype.beartype` decorator).
        For efficiency, this code object should *always* be accessed in lieu of
        inefficiently calling the comparatively slower
        :func:`beartype._util.func.utilfunccodeobj.get_func_codeobj` getter.
    func_wrapper : Callable
        **Wrapper callable** to be unwrapped in the event that the
        :attr:`func_wrappee` differs from the callable to be unwrapped.
        Typically, these two callables are the same. Edge cases in which these
        two callables differ include:

        * When the wrapper callable is a **pseudo-callable** (i.e., otherwise
          uncallable object whose type renders that object callable by defining
          the ``__call__()`` dunder method) *and* the :attr:`func_wrappee` is
          the ``__call__()`` dunder method. If that pseudo-callable wraps a
          lower-level callable, then that pseudo-callable (rather than that
          ``__call__()`` dunder method) defines the ``__wrapped__`` instance
          variable providing that callable.

        This callable is typically identical to the :attr:`func_wrappee`.
    func_wrapper_code_call_prefix : str
        Code snippet prefixing all calls to the decorated callable in the body
        of the wrapper function wrapping that callable with type checking. This
        string is guaranteed to be either:

        * If the decorated callable is synchronous (i.e., neither a coroutine
          nor asynchronous generator), the empty string.
        * If the decorated callable is asynchronous (i.e., either a coroutine
          nor asynchronous generator), the ``"await "`` keyword.
    func_wrapper_code_return_prefix : str
        Code snippet prefixing the value to be returned or yielded to the caller
        after calling the decorated callable in the body of the wrapper function
        wrapping that callable with type checking. This string is guaranteed to
        be either:

        * If the decorated callable is a :pep:`380`-compliant synchronous
          generator factory, ``"yield from "``.
        * If the decorated callable is a :pep:`525`-compliant asynchronous
          generator factory, a pure-Python snippet safely implementing the
          hypothetical equivalent of an ``"async yield from "`` expression (if
          that expression existed, which it doesn't).
        * Else, ``"return "``.
    func_wrapper_code_signature_prefix : str
        Code snippet prefixing the signature declaring the wrapper function
        wrapping the decorated callable with type checking. This string is
        guaranteed to be either:

        * If the decorated callable is synchronous (i.e., neither a coroutine
          nor asynchronous generator), the empty string.
        * If the decorated callable is asynchronous (i.e., either a coroutine
          or asynchronous generator), the ``"async "`` keyword.
    func_wrapper_name : str
        Unqualified basename of the type-checking wrapper function to be
        generated and returned by the current invocation of the
        :func:`beartype.beartype` decorator.
    func_wrapper_scope : LexicalScope
        **Local scope** (i.e., dictionary mapping from the name to value of
        each attribute referenced in the signature) of this wrapper function.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'cls_stack',
        'conf',
        'func_annotations',
        'func_annotations_get',
        '_is_func_annotations_dirty',
        'func_wrappee',
        'func_wrappee_is_nested',
        'func_wrappee_scope_forward',
        'func_wrappee_scope_nested_names',
        'func_wrappee_wrappee',
        'func_wrappee_wrappee_codeobj',
        'func_wrapper',
        'func_wrapper_code_call_prefix',
        'func_wrapper_code_return_prefix',
        'func_wrapper_code_signature_prefix',
        'func_wrapper_name',
        'func_wrapper_scope',
    )

    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        cls_stack: TypeStack
        conf: BeartypeConf
        func_annotations: Pep649HintableAnnotations
        func_annotations_get: Callable[[str, object], object]
        _is_func_annotations_dirty: bool
        func_wrappee: Callable
        func_wrappee_is_nested: bool
        func_wrappee_scope_forward: Optional[BeartypeForwardScope]
        func_wrappee_scope_nested_names: Optional[FrozenSet[str]]
        func_wrappee_wrappee: Callable
        func_wrappee_wrappee_codeobj: CallableCodeObjectType
        func_wrapper: Callable
        func_wrapper_code_call_prefix: str
        func_wrapper_code_return_prefix: str
        func_wrapper_code_signature_prefix: str
        func_wrapper_name: str
        func_wrapper_scope: LexicalScope

    # Coerce instances of this class to be unhashable, preventing spurious
    # issues when accidentally passing these instances to memoized callables by
    # implicitly raising a "TypeError" exception on the first call to those
    # callables. There exists no tangible benefit to permitting these instances
    # to be hashed (and thus also cached), since these instances are:
    # * Specific to the decorated callable and thus *NOT* safely cacheable
    #   across functions applying to different decorated callables.
    # * Already cached via the acquire_instance() function called by the
    #   "beartype._decor.decormain" submodule.
    #
    # See also:
    #     https://docs.python.org/3/reference/datamodel.html#object.__hash__
    __hash__ = None  # type: ignore[assignment]

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this metadata by nullifying all instance variables.

        Caveats
        -------
        **This class is not intended to be explicitly instantiated.** Instead,
        callers are expected to (in order):

        #. Acquire cached instances of this class via the
           :mod:`beartype._util.cache.pool.utilcachepoolinstance` submodule.
        #. Call the :meth:`reinit` method on these instances to properly
           initialize these instances.
        '''

        # Nullify instance variables for safety.
        self.deinit()


    def deinit(self) -> None:
        '''
        Deassociate this metadata from the callable passed to the most recent
        call of the :meth:`reinit` method, typically before releasing this
        instance of this class back to the
        :mod:`beartype._util.cache.pool.utilcachepoolobject` submodule.

        This method prevents a minor (albeit still undesirable, of course)
        memory leak in which this instance would continue to remain accidentally
        associated with that callable despite this instance being released back
        to its object pool, which would then prevent that callable from being
        garbage-collected on the finalization of the last external reference to
        that callable.
        '''

        # Restore instance variables to initial defaults.
        self.func_wrapper_scope: LexicalScope = {}
        self._is_func_annotations_dirty = False

        # Nullify all remaining instance variables for safety.
        self.cls_stack = (  # type: ignore[assignment]
        self.conf) = (  # type: ignore[assignment]
        self.func_annotations) = (  # type: ignore[assignment]
        self.func_annotations_get) = (  # type: ignore[assignment]
        self.func_wrappee) = (  # type: ignore[assignment]
        self.func_wrappee_is_nested) = (  # type: ignore[assignment]
        self.func_wrappee_scope_forward) = (  # type: ignore[assignment]
        self.func_wrappee_scope_nested_names) = (  # type: ignore[assignment]
        self.func_wrappee_wrappee) = (  # type: ignore[assignment]
        self.func_wrappee_wrappee_codeobj) = (  # type: ignore[assignment]
        self.func_wrapper) = (  # type: ignore[assignment]
        self.func_wrapper_code_call_prefix) = (  # type: ignore[assignment]
        self.func_wrapper_code_return_prefix) = (  # type: ignore[assignment]
        self.func_wrapper_code_signature_prefix) = (  # type: ignore[assignment]
        self.func_wrapper_name) = None  # type: ignore[assignment]


    def reinit(
        self,

        # Mandatory parameters.
        func: Callable,
        conf: BeartypeConf,

        # Optional parameters.
        cls_stack: TypeStack = None,
        wrapper: Optional[Callable] = None,
    ) -> None:
        '''
        Reinitialize this metadata from the passed callable, typically after
        acquisition of a previously cached instance of this class from the
        :mod:`beartype._util.cache.pool.utilcachepoolobject` submodule.

        If :pep:`563` is conditionally active for this callable, this function
        additionally resolves all postponed annotations on this callable to
        their referents (i.e., the intended annotations to which those
        postponed annotations refer).

        Parameters
        ----------
        func : Callable
            Callable currently being decorated by :func:`beartype.beartype`.
        conf : BeartypeConf
            Beartype configuration configuring :func:`beartype.beartype`
            specific to this callable.
        cls_stack : TypeStack
            **Type stack** (i.e., either tuple of zero or more arbitrary types
            *or* :data:`None`). See also the parameter of the same name accepted
            by the :func:`beartype._decor.decorcore.beartype_object` function.
        wrapper : Optional[Callable]
            **Wrapper callable** to be unwrapped in the event that the callable
            currently being decorated by :func:`beartype.beartype` differs from
            the callable to be unwrapped. Typically, these two callables are the
            same. Edge cases in which these two callables differ include:

            * When ``wrapper`` is a **pseudo-callable** (i.e., otherwise
              uncallable object whose type renders that object callable by
              defining the ``__call__()`` dunder method) *and* ``func`` is that
              ``__call__()`` dunder method. If that pseudo-callable wraps a
              lower-level callable, then that pseudo-callable (rather than that
              ``__call__()`` dunder method) defines the ``__wrapped__`` instance
              variable providing that callable.

            Defaults to :data:`None`, in which case this parameter *actually*
            defaults to ``func``.

        Raises
        ------
        BeartypePep563Exception
            If evaluating a postponed annotation on this callable raises an
            exception (e.g., due to that annotation referring to local state no
            longer accessible from this deferred evaluation).
        BeartypeDecorWrappeeException
            If either:

            * This callable is uncallable.
            * This callable is neither a pure-Python function *nor* method;
              equivalently, if this callable is either C-based *or* a class or
              object defining the ``__call__()`` dunder method.
            * This configuration is *not* actually a configuration.
            * ``cls_owner`` is neither a class *nor* :data:`None`.
        '''

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # CAUTION: Note this method intentionally avoids creating and passing an
        # "exception_prefix" substring to callables called below. Why? Because
        # exhaustive profiling has shown that creating that substring consumes a
        # non-trivial slice of decoration time. In other words, raw efficiency.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        # ..................{ VALIDATE                       }..................
        # If the caller failed to pass a callable to be unwrapped, default that
        # to the callable to be type-checked.
        if wrapper is None:
            wrapper = func
        # Else, the caller passed a callable to be unwrapped. Preserve it up!
        # print(f'Beartyping func {repr(func)} + wrapper {repr(wrapper)}...')

        # If the callable to be type-checked is uncallable, raise an exception.
        if not callable(func):
            raise BeartypeDecorWrappeeException(f'{repr(func)} uncallable.')
        # Else, that callable is callable.
        #
        # If the callable to be unwrapped is uncallable, raise an exception.
        elif not callable(wrapper):
            raise BeartypeDecorWrappeeException(f'{repr(wrapper)} uncallable.')
        # Else, that callable is callable.
        #
        # If this configuration is *NOT* a configuration, raise an exception.
        elif not isinstance(conf, BeartypeConf):
            raise BeartypeDecorWrappeeException(
                f'BeartypeDecorMeta.reinit() method "conf" parameter '
                f'{repr(conf)} not beartype configuration.'
            )
        # Else, this configuration is a configuration.
        #
        # If this class stack is neither a tuple *NOR* "None", raise an
        # exception.
        elif not isinstance(cls_stack, NoneTypeOr[tuple]):
            raise BeartypeDecorWrappeeException(
                f'BeartypeDecorMeta.reinit() method "cls_stack" parameter '
                f'{repr(cls_stack)} neither tuple nor "None".'
            )
        # Else, this class stack is either a tuple *OR* "None".

        # If the caller passed a non-empty class stack...
        if cls_stack:
            # For each item of this class stack...
            for cls_stack_item in cls_stack:
                # If this item is *NOT* a type, raise an exception.
                if not isinstance(cls_stack_item, type):
                    raise BeartypeDecorWrappeeException(
                        f'BeartypeDecorMeta.reinit() method "cls_stack" item '
                        f'{repr(cls_stack_item)} not type.'
                    )
                # Else, this item is a type.
        # Else, the caller either passed no class stack *OR* an empty class
        # stack. In either case, ignore this parameter.

        # ..................{ VARS                           }..................
        # Classify all passed parameters.
        self.conf = conf
        self.cls_stack = cls_stack

        # ..................{ VARS ~ func : wrappee          }..................
        # Wrappee callable currently being decorated.
        self.func_wrappee = func

        # True only if this wrappee callable is nested. As a minor efficiency
        # gain, we can avoid the slightly expensive call to is_func_nested() by
        # noting that:
        # * If the class stack is non-empty, then this wrappee callable is
        #   necessarily nested in one or more classes.
        # * Else, defer to the is_func_nested() tester.
        self.func_wrappee_is_nested = bool(cls_stack) or is_func_nested(func)

        # Defer the resolution of both global and local scopes for this wrappee
        # callable until needed to subsequently resolve stringified type hints.
        self.func_wrappee_scope_forward = None
        self.func_wrappee_scope_nested_names = None

        # Possibly wrapped wrappee code object (i.e., code object underlying the
        # callable currently being type-checked by the @beartype decorator) if
        # this wrappee is pure-Python *OR* "None" otherwise.
        #
        # Note that only the possibly unwrapped wrappee wrappee defined below
        # (i.e., "func_wrappee_wrappee") *MUST* be pure-Python and thus *MUST*
        # have a code object. This higher-level wrappee is permitted to be
        # C-based and thus need *NOT* have a code object.
        func_wrappee_codeobj = get_func_codeobj_or_none(func)

        # ..................{ VARS ~ func : wrappee wrappee  }..................
        # Possibly unwrapped callable unwrapped from this wrappee callable.
        self.func_wrappee_wrappee = unwrap_func_all_isomorphic(
            func=func, wrapper=wrapper)
        # print(f'func_wrappee: {self.func_wrappee}')
        # print(f'func_wrappee_wrappee: {self.func_wrappee_wrappee}')
        # print(f'{dir(self.func_wrappee_wrappee)}')

        # Possibly unwrapped callable code object.
        self.func_wrappee_wrappee_codeobj = get_func_codeobj(
            func=self.func_wrappee_wrappee,
            exception_cls=BeartypeDecorWrappeeException,
        )

        # ..................{ VARS ~ func : wrapper          }..................
        # Wrapper callable to be unwrapped in the event that the
        # decorated callable differs from the callable to be unwrapped.
        self.func_wrapper = wrapper

        # Efficiently reduce this local scope back to the dictionary of all
        # parameters unconditionally required by *ALL* wrapper functions.
        self.func_wrapper_scope.clear()

        # Machine-readable name of the wrapper function to be generated.
        self.func_wrapper_name = func.__name__

        # ..................{ VARS ~ func : hints            }..................
        # Dictionary mapping from the name of each annotated parameter accepted
        # by the unwrapped callable to the type hint annotating that parameter
        # *AFTER* resolving all postponed type hints elsewhere.
        #
        # Note that:
        # * The functools.update_wrapper() function underlying the
        #   @functools.wrap decorator underlying all sane decorators propagates
        #   this dictionary from lower-level wrappees to higher-level wrappers
        #   by default. We intentionally classify the annotations dictionary of
        #   this higher-level wrapper, which *SHOULD* be the superset of that of
        #   this lower-level wrappee (and thus more reflective of reality).
        # * The type hints annotating the callable to be unwrapped (i.e.,
        #   "wrapper)" are preferred to those annotating the callable to be
        #   type-checked (i.e., "func"). Why? Because the callable to be
        #   unwrapped is either the original pure-Python function or method
        #   defined by the user *OR* a pseudo-callable object transitively
        #   wrapping that function or method; in either case, the type hints
        #   annotating that callable are guaranteed to be authoritative.
        #   However, the callable to be type-checked is in this case typically
        #   only a thin isomorphic wrapper deferring to the callable to be
        #   unwrapped.
        #
        # Consider the typical use case invoking this conditional logic:
        #     from functools import update_wrapper, wraps
        #
        #     def probably_lies(lies: str, more_lies: str) -> str:
        #         return lies + more_lies
        #
        #     class LyingClass(object):
        #         def __call__(self, *args, **kwargs):
        #             return probably_lies(*args, **kwargs)
        #
        #     cheating_object = LyingClass()
        #     update_wrapper(wrapper=cheating_object, wrapped=probably_lies)
        #     print(cheating_object.__annotations__)
        #
        # ...which would print:
        #     {'lies': <class 'str'>, 'more_lies': <class 'str'>, 'return':
        #     <class 'str'>}
        #
        # We thus see that this use case successfully propagated the
        # "__annotations__" dunder dictionary from the probably_lies()
        # function onto the pseudo-callable "cheating_object" object.
        #
        # In this case, the caller would have called this method as:
        #     decor_meta.reinit(
        #         func=cheating_object.__call__, wrapper=cheating_object)
        #
        # Note that "func" (i.e., the callable to be type-checked) is only a
        # thin isomorphic wrapper deferring to "wrapper" (i.e., the callable to
        # be unwrapped). Even if "func" were annotated with type hints, those
        # type hints would be useless for most intents and purposes.
        self.func_annotations = get_pep649_hintable_annotations(
            hintable=wrapper, exception_cls=BeartypeDecorWrappeeException)
        # print(f'Beartyping func {repr(func)} + wrapper {repr(wrapper)} w/ annotations {self.func_annotations}...')

        # dict.get() method bound to this dictionary.
        self.func_annotations_get = self.func_annotations.get

        # ..................{ VARS ~ func : kind             }..................
        # Default the return prefix to the "return" keyword, which suffices to
        # return the type-checked value returned by this wrappee for all kinds
        # of wrappees *EXCEPT* generator factories. (See below.)
        self.func_wrapper_code_return_prefix = 'return '

        # Default all remaining code snippets to the empty string.
        self.func_wrapper_code_signature_prefix = ''
        self.func_wrapper_code_call_prefix = ''

        # If the wrappee callable currently being decorated is pure-Python...
        if func_wrappee_codeobj:
            # print(f'Decorated callable {repr(func)}...')
            # print(f'synchronous generator factory? {is_func_sync_generator(func_wrappee_codeobj)}')

            # If this wrappee is an asynchronous coroutine factory (i.e.,
            # callable declared with "async def" rather than merely "def"
            # keywords and containing *NO* "yield" expressions)...
            #
            # Note that:
            # * The code object of the higher-level wrapper rather than
            #   lower-level wrappee is passed. Why? Because @beartype directly
            #   decorates *ONLY* the former, whose asynchronicity has *NO*
            #   relation to that of the latter. Notably, it is both feasible and
            #   (relatively) commonplace for third-party decorators to enable:
            #   * Synchronous callables to be called asynchronously by wrapping
            #     synchronous callables with asynchronous closures.
            #   * Asynchronous callables to be called synchronously by wrapping
            #     asynchronous callables with synchronous closures. Indeed, our
            #     top-level "conftest.py" pytest plugin does exactly this --
            #     enabling asynchronous tests to be safely called by pytest's
            #     currently synchronous framework.
            # * The higher-level is_func_async() tester is intentionally *NOT*
            #   called here, as doing so would also implicitly prefix all calls
            #   to asynchronous generator factories (i.e., callables also
            #   declared with the "async def" rather than merely "def" keywords
            #   but containing one or more "yield" expressions) with the "await"
            #   keyword. Whereas asynchronous coroutine objects implicitly
            #   returned by all asynchronous coroutine callables return a single
            #   awaitable value, asynchronous generator objects implicitly
            #   returned by all asynchronous generator callables *NEVER* return
            #   any awaitable value; they instead yield one or more values to
            #   external "async for" loops.
            if is_func_coro(func_wrappee_codeobj):
                # Code snippet prefixing the declaration of the wrapper function
                # wrapping this coroutine factory with type-checking.
                self.func_wrapper_code_signature_prefix = 'async '

                # Code snippet prefixing all calls to this coroutine factory.
                self.func_wrapper_code_call_prefix = 'await '
            # Else, this wrappee is *NOT* an asynchronous coroutine factory.
            #
            # If this wrappee is a synchronous generator factory (i.e.,
            # callable declared with merely the "def" rather than "async def"
            # keywords and containing one or more "yield" expressions)...
            elif is_func_sync_generator(func_wrappee_codeobj):
                # print(f'Decorated synchronous generator factory {repr(func)} detected!')

                # Code snippet yielding from (i.e., deferring to) the
                # synchronous generator returned by calling this PEP
                # 380-compliant synchronous generator factory.
                #
                # Note that:
                # * This prefix is required *ONLY* to force the active Python
                #   interpreter to implicitly enable the "inspect.CO_GENERATOR"
                #   bit flag in the "func.__code__.co_flags" bit field of the
                #   high-level wrapper subsequently generated by the @beartype
                #   decorator wrapping this lower-level wrappee with
                #   type-checking when that wrapper is defined. Why? Because the
                #   standard inspect.isgeneratorfunction() tester conflates a
                #   fictitious one-to-one relation between synchronous generator
                #   factories and pure-Python callables enabling this bit flag.
                #   If @beartype neglected to add "yield from" expressions to
                #   the code generated for type-checking wrappers wrapping
                #   synchronous generator factories, then:
                #   * inspect.isgeneratorfunction() would declare those wrappers
                #     to *NOT* be synchronous generator factories despite those
                #     wrappers literally being such factories.
                #   * Third-party packages (e.g., Gradio) improperly calling
                #     inspect.isgeneratorfunction() to detect synchronous
                #     generator factories would misidentify those wrappers to
                #     *NOT* be synchronous generator factories.
                # * Technically, high-level synchronous generator factories that
                #   defer to lower-level synchronous generator factories do
                #   *NOT* need to perform "yield from" expressions to be usable
                #   as synchronous generator factories. Ergo, the standard
                #   inspect.isgeneratorfunction() tester blatantly returns false
                #   negatives in common use cases. Previously, @beartype was
                #   such a use case. No one should call that tester. Of course,
                #   everyone calls that tester.
                # * This prefix serves *NO* other practical or useful purpose.
                #   Indeed, @beartype previously generated high-level wrappers
                #   for synchronous generator factories simply by calling those
                #   factories and returning the synchronous generators returned
                #   by those calls. That worked for a decade... until
                #   third-party packages outside our control began
                #   inappropriately calling inspect.isgeneratorfunction().
                #
                # See also this closed issue:
                #     https://github.com/beartype/beartype/issues/423

                #FIXME: Actually, this has an additional side benefit. Thanks to
                #this, we can now type-check the value returned by a generator
                #as well against the child hint "{hint_return}" if the return
                #hint annotating this generator factory is of the expanded form
                #"Generator[{hint_yield}, {hint_send}, {hint_return}]". Of
                #course, basically *NOBODY* ever returns *ANYTHING* from a
                #generator. So, it's unclear how many practical value this has.
                #Still, it's feasible now. And that's half the battle.
                #FIXME: Actually, we can do *FAR* better than that. By employing
                #a similar technique as the solution below for asynchronous
                #generators, we can now efficiently (and, more importantly,
                #safely) type-check both synchronous and asynchronous generator
                #*YIELDS* in O(1) time. How? Just:
                #* If this callable is decorated by a return hint (which is
                #  trivially detectable here) *AND* the yield-specific child
                #  hint subscripting that return hint is unignorable (e.g., is
                #  neither "object" nor "typing.Any"), then we should instead
                #  assign "self.func_wrapper_code_return_prefix" here to a
                #  full-blown code snippet implementing our best approximation
                #  of "yield from" in pure-Python augmented with type-checking.
                #* Else, preserve this assignment of
                #  "self.func_wrapper_code_return_prefix" to "yield from", which
                #  is both implemented in C *AND* is the official
                #  implementation. So, it's guaranteed to be the fastest *AND*
                #  safest means of deferring between synchronous generators.

                self.func_wrapper_code_return_prefix = 'yield from '
            # Else, this wrappee is *NOT* a synchronous generator factory.
            #
            # If this wrappee is a asynchronous generator factory (i.e.,
            # callable declared with "async def" keywords and containing one or
            # more "yield" expressions)...
            elif is_func_async_generator(func_wrappee_codeobj):
                # Code snippet prefixing the declaration of the wrapper function
                # wrapping this coroutine factory with type-checking.
                self.func_wrapper_code_signature_prefix = 'async '

                # Code snippet yielding from (i.e., deferring to) the
                # asynchronous generator returned by calling this PEP
                # 525-compliant asynchronous generator factory.
                #
                # Note that:
                # * All of the caveats documented above for the case in which
                #   this wrappee is a synchronous generator factory also apply
                #   here, substituting:
                #   * "inspect.CO_GENERATOR" with "CO_ASYNC_GENERATOR".
                #   * inspect.isgeneratorfunction() with
                #     inspect.isasyncgenfunction().
                # * Additional caveats also apply. Currently, there exists *NO*
                #   C-based "async yield from" expression for asynchronous
                #   generators equivalent to the PEP 380-compliant C-based
                #   "yield from" expression for synchronous generators. This
                #   pure-Python code snippet is our best approximation of a
                #   pure-Python "async yield from" expression. See the docstring
                #   and commentary surrounding the definition of this snippet
                #   for gruesome details that will exhaust your will to code.
                #
                # See also this closed issue:
                #     https://github.com/beartype/beartype/issues/592
                self.func_wrapper_code_return_prefix = CODE_PEP525_RETURN_PREFIX
            # Else, this wrappee is *NOT* an asynchronous generator factory.
            #
            # In this case, preserve these code snippets as the empty string.
        # Else, this wrappee is *NOT* pure-Python. In this case, preserve these
        # code snippets as the empty string.

    # ..................{ DUNDERS                            }..................
    def __repr__(self) -> str:
        '''
        Machine-readable representation of this metadata.
        '''

        # Represent this metadata with just the minimal subset of metadata
        # needed to reasonably describe this metadata.
        return (
            f'{self.__class__.__name__}('
            f'func={repr(self.func_wrappee)}, '
            f'conf={repr(self.conf)}'
            f')'
        )

    # ..................{ SETTERS                            }..................
    #FIXME: Unit test us up, please.
    def set_func_annotations_if_dirty(self) -> None:
        '''
        Safely set the ``__annotations__`` dunder dictionary annotating the
        decorated callable to the possibly mutated type hint dictionary
        associated with that callable if the latter is **dirty** (i.e., modified
        from the original type hint dictionary annotating the decorated callable
        by a prior call to the :meth:`set_func_pith_hint` setter).
        '''

        # If the type hint dictionary associated with the decorated callable is
        # dirty (i.e., changed from the original "__annotations__" dunder
        # dictionary annotating that callable), register these changes in a
        # manner compliant with both PEP 649 and Python >= 3.14. Specifically...
        if self._is_func_annotations_dirty:
            # print(f'Setting new function {self.func_wrapper} annotations {self.func_annotations}...')

            # Set these modified annotations on the decorated callable
            # originating these annotations.
            set_pep649_hintable_annotations(
                hintable=self.func_wrapper, annotations=self.func_annotations)

            # Note this type hint dictionary to now be clean.
            self._is_func_annotations_dirty = False
        # Else, the type hint dictionary associated with the decorated callable
        # is still clean (i.e., identical to the original "__annotations__"
        # dunder dictionary annotating that callable). In this case, silently
        # reduce to a noop.


    #FIXME: Unit test us up, please.
    def set_func_pith_hint(self, pith_name: str, hint: Hint) -> None:
        '''
        Safely set the hint annotating the parameter or return with the passed
        name of the decorated callable to the passed hint in a portable manner
        consistent with Python >= 3.14 (and thus both :pep:`649` and :pep:`749`,
        which fundamentally complicate hint mutation).

        Parameters
        ----------
        pith_name : str
            Name of the parameter or return to be re-annotated.
        hint: Hint
            Hint to re-annotate this parameter or return to.

        Raises
        ------
        BeartypeDecorWrappeeException
            If the decorated callable accepts *no* parameter or return with the
            passed name.
        '''
        assert isinstance(pith_name, str), f'{repr(pith_name)} not string.'
        # print(f'Setting new function {self.func_wrapper} pith "{pith_name}" hint {hint}...')

        # If the decorated callable accepts *NO* parameter or return with the
        # passed name, raise an exception.
        if pith_name not in self.func_annotations:
            raise BeartypeDecorWrappeeException(
                f'{prefix_callable_pith(func=self.func_wrapper, pith_name=pith_name)}'
                f'unrecognized.'
            )
        # Else, the decorated callable accepts this parameter or return.

        # Note that the type hint dictionary is now dirty (i.e., modified from
        # the original type hint dictionary annotating the decorated callable).
        self._is_func_annotations_dirty = True

        # Set the hint annotating this parameter or return to the passed hint.
        self.func_annotations[pith_name] = hint

    # ..................{ LABELLERS                          }..................
    def label_func_wrapper(self) -> str:
        '''
        Human-readable label describing the type-checking wrapper function to be
        generated and returned by the current invocation of the
        :func:`beartype.beartype` decorator.

        This method is a non-negligible optimization. Since string munging is
        *extremely* slow and this method necessarily munges strings, external
        callers delay this string munging as late as possible by delaying all
        calls to this method as late as possible (e.g., until an exception
        message requiring this label is actually required).
        '''

        # One-liner of Ultimate Beauty: we invoke thee in this line!
        return f'@beartyped {self.func_wrapper_name}() wrapper'

# ....................{ FACTORIES                          }....................
#FIXME: Unit test us up, please.
def make_beartype_call(**kwargs) -> BeartypeDecorMeta:
    '''
    **Beartype call metadata** (i.e., object encapsulating *all* metadata for
    the passed user-defined callable, typically currently being decorated by the
    :func:`beartype.beartype` decorator).

    Caveats
    -------
    **This higher-level factory function should always be called in lieu of
    instantiating the** :class:`.BeartypeDecorMeta` **class directly.** Why?
    Brute-force efficiency. This factory efficiently reuses previously
    instantiated :class:`.BeartypeDecorMeta` objects rather than inefficiently
    instantiating new :class:`.BeartypeDecorMeta` objects.

    **The caller must pass the metadata returned by this factory back to the**
    :func:`beartype._util.cache.pool.utilcachepoolinstance.release_instance`
    **function.** If accidentally omitted, this metadata will simply be
    garbage-collected rather than available for efficient reuse by this factory.
    Although hardly a worst-case outcome, omitting that explicit call largely
    defeats the purpose of calling this factory in the first place.

    Parameters
    ----------
    All keyword parameters are passed as is to the :meth:`.BeartypeDecorMeta.reinit`
    method.

    Returns
    -------
    BeartypeDecorMeta
        Beartype call metadata describing this callable.
    '''

    # Acquire previously cached beartype call metadata from its object pool.
    decor_meta = acquire_instance(BeartypeDecorMeta)

    # Reinitialize this metadata with the passed keyword parameters.
    decor_meta.reinit(**kwargs)

    # Return this metadata.
    return decor_meta


#FIXME: Unit test us up, please.
def cull_beartype_call(decor_meta: BeartypeDecorMeta) -> None:
    '''
    Deinitialize the passed **beartype call metadata** (i.e., object
    encapsulating *all* metadata for the passed user-defined callable, typically
    currently being decorated by the :func:`beartype.beartype` decorator).

    Parameters
    ----------
    decor_meta : BeartypeDecorMeta
        Beartype call metadata to be deinitialized.
    '''

    # If the type hint dictionary associated with the decorated callable is
    # dirty (i.e., changed from the original "__annotations__" dunder dictionary
    # annotating that callable), register these changes in a manner compliant
    # with both PEP 649 and Python >= 3.14 *BEFORE* nullifying these type hints.
    decor_meta.set_func_annotations_if_dirty()

    # Deinitialize this beartype call metadata.
    decor_meta.deinit()

    # Release this beartype call metadata back to its object pool.
    release_instance(decor_meta)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **nuitka integration tests.**

This submodule functionally tests that the third-party ``nuitka`` compiler
successfully compiles this pure-Python project.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import (
    skip_unless_os_linux,
    skip_unless_package,
    skip_unless_pathable,
)

# ....................{ TESTS                              }....................
# Skip this "nuitka"-specific functional test unless all of the following apply:
# * The "nuitka" package is importable under the active Python interpreter.
# * The third-party GCC compiler is in the current ${PATH}.
# * The current platform is *NOT* a Linux distribution. Although "nuitka"
#   technically *CAN* be used under non-Linux platforms, doing so is typically
#   non-trivial and likely to unexpectedly explode with catastrophic errors.
@skip_unless_os_linux()
@skip_unless_package('nuitka')
@skip_unless_pathable('gcc')
def test_nuitka(capsys, tmp_path: 'pathlib.Path') -> None:
    '''
    Integration test testing that the third-party ``nuitka`` compiler
    successfully compiles a minimal-length example (MLE) extensively leveraging
    this project.

    Parameters
    ----------
    capsys
        :mod:`pytest`-specific systems capability fixture.
    tmp_path : pathlib.Path
        Abstract path encapsulating a temporary directory unique to this test,
        created in the base temporary directory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.py.utilpyinterpreter import (
        get_interpreter_command_words)
    from beartype_test._util.command.pytcmdrun import (
        run_command_forward_output,
        run_command_forward_stderr_return_stdout,
    )
    from beartype_test._util.os.pytosshell import shell_quote
    from beartype_test._util.path.pytpathtest import (
        get_test_func_data_lib_nuitka_file)

    # ....................{ COMMAND                        }....................
    #FIXME: "nuitka" occasionally attempts to download external third-party
    #tooling. By default, it interactively prompts the user before doing so.
    #This is obviously bad for automated testing. Thankfully, "nuitka" avoids
    #doing so when it detects that stdin has been redirected to "/dev/null".
    #Ergo, improve the run_command_forward_output() function called below to
    #*ALWAYS* redirect stdin to "/dev/null". We *ALWAYS* want that everywhere.

    # Tuple of all shell words with which to run the external "nuitka" command.
    NUITKA_ARGS = get_interpreter_command_words() + (
        # Fully-qualified name of the "nuitka" package to be run.
        '-m', 'nuitka',

        # Do *NOT* attempt to cache the compilation of intermediate C artifacts
        # with the third-party "ccache" command, which may not even necessarily
        # be installed into the current ${PATH}.
        '--disable-ccache',

        #FIXME: Currently disabled as enabling even just this incurs *FAR* too
        #much time and space. Sadness ensues.
        # # Instruct "nuitka" to additionally compile the entirety of the
        # # "beartype" package. By default, "nuitka" only compiles the script
        # # (specified below).
        # '--follow-import-to=beartype',
 
        # Absolute or relative dirname of a test-specific temporary directory to
        # which "nuitka" will generate ignorable compilation artifacts.
        f'--output-dir={shell_quote(str(tmp_path))}',

        # Absolute filename of a minimal-length example (MLE) leveraging this
        # project to be compiled by "nuitka".
        str(get_test_func_data_lib_nuitka_file()),
    )

    # With pytest's default capturing of standard output and error temporarily
    # disabled...
    #
    # Technically, this is optional. Pragmatically, "nuitka" is sufficiently
    # slow that failing to do this renders this test silent for several tens of
    # seconds to minutes, which is significantly worse than printing progress.
    with capsys.disabled():
        # Run the "nuitka" command in the current ${PATH} with these options and
        # arguments, raising an exception on subprocess failure while forwarding
        # all standard output and error output by this subprocess to the
        # standard output and error file handles of the active Python process.
        run_command_forward_output(command_words=NUITKA_ARGS)

    # ....................{ ASSERT                         }....................
    # Absolute or relative filename of the executable binary generated by
    # "nuitka" after running the above command.
    COMPILED_FILENAME = str(tmp_path / 'beartype_nuitka.bin')

    # Standard output emitted by running this executable binary.
    COMPILED_STDOUT = run_command_forward_stderr_return_stdout(
        COMPILED_FILENAME)

    # Assert that this is the expected output.
    assert COMPILED_STDOUT == (
'''ClassTypeHint(<class 'int'>)
ClassTypeHint(<class 'float'>)''')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator PEP-noncompliant type hint unit tests.**

This submodule unit tests the :func:`beartype.beartype` decorator with respect
to **PEP-noncompliant type hints** (i.e., :mod:`beartype`-specific annotations
*not* compliant with annotation-centric PEPs).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TODO                               }....................
#FIXME: Define new unit tests exercising:
#* Python >= 3.8-specific keyword-only parameters.

# ....................{ TESTS ~ pass : param : kind        }....................
def test_nonpep_param_kind_positional_or_keyword_pass() -> None:
    '''
    Test successful usage of the :func:`beartype.beartype` decorator for a
    function call passed non-variadic positional and/or keyword parameters
    annotated with PEP-noncompliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype

    # Decorated callable to be exercised.
    @beartype
    def slaanesh(daemonette: str, keeper_of_secrets: str) -> str:
        return daemonette + keeper_of_secrets

    # Assert that calling this callable with both positional and keyword
    # arguments returns the expected return value.
    assert slaanesh(
        'Seeker of Decadence', keeper_of_secrets="N'Kari") == (
        "Seeker of DecadenceN'Kari")


def test_nonpep_param_kind_variadic_and_keyword_only_pass() -> None:
    '''
    Test successful usage of the :func:`beartype.beartype` decorator for a
    function call passed variadic positional parameters followed by a
    keyword-only parameter, all annotated with PEP-noncompliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype

    # Decorated callable to be exercised.
    @beartype
    def changer_of_ways(
        sky_shark: str, *dark_chronology: int, chaos_spawn: str) -> str:
        return (
            sky_shark +
            str(dark_chronology[0]) +
            str(dark_chronology[-1]) +
            chaos_spawn
        )

    # Assert that calling this callable with variadic positional parameters
    # followed by a keyword-only parameter returns the expected return value.
    assert changer_of_ways(
        'Screamers', 0, 1, 15, 25, chaos_spawn="Mith'an'driarkh") == (
        "Screamers025Mith'an'driarkh")



def test_nonpep_param_kind_variadic_fail() -> None:
    '''
    Test unsuccessful usage of the :func:`beartype.beartype` decorator for a
    function call passed variadic positional parameters annotated with
    PEP-noncompliant type hints.
    '''

    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from pytest import raises

    # Decorated callable to be exercised.
    @beartype
    def imperium_of_man(
        space_marines: str, *ceaseless_years: int, primarch: str) -> str:
        return space_marines + str(ceaseless_years[1]) + primarch

    # Assert that calling this callable with invalid variadic positional
    # parameters raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        imperium_of_man(
            'Legiones Astartes', 30, 31, 36, 'M41', primarch='Leman Russ')

# ....................{ TESTS ~ pass : param : type        }....................
def test_nonpep_pass_param_tuple() -> None:
    '''
    Test type-checking for a function call successfully passed a parameter
    annotated with a PEP-noncompliant tuple union.
    '''

    # Import this decorator.
    from beartype import beartype

    # Function to be type-checked. For completeness, test both an actual class
    # *AND* a forward reference to an actual class in this tuple annotation.
    @beartype
    def genestealer(tyranid: str, hive_fleet: (str, 'builtins.int')) -> str:
        return tyranid + str(hive_fleet)

    # Call this function with each of the two types listed in the above tuple.
    assert genestealer('Norn-Queen', 'Behemoth') == 'Norn-QueenBehemoth'
    assert genestealer('Carnifex', 0xDEADBEEF) == 'Carnifex3735928559'


def test_nonpep_pass_param_custom() -> None:
    '''
    Test type-checking for a function call successfully passed a parameter
    annotated as a user-defined class.
    '''

    # Import this decorator.
    from beartype import beartype

    # User-defined type.
    class CustomTestStr(str):
        pass

    # Function to be type-checked.
    @beartype
    def hrud(gugann: str, delphic_plague: CustomTestStr) -> str:
        return gugann + delphic_plague

    # Call this function with each of the above type.
    assert hrud(
        'Troglydium hruddi', delphic_plague=CustomTestStr('Delphic Sink')) == (
        'Troglydium hruddiDelphic Sink')

# ....................{ TESTS ~ fail : param : call        }....................
def test_nonpep_fail_param_call_tuple() -> None:
    '''
    Test type-checking for a function call unsuccessfully passed a parameter
    annotated with a PEP-noncompliant tuple union.
    '''

    # Import this decorator.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from pytest import raises

    # Annotated function to be type-checked.
    @beartype
    def eldar(isha: str, asuryan: (str, int)) -> str:
        return isha + asuryan

    # Call this function with an invalid type and assert the expected
    # exception.
    with raises(BeartypeCallHintParamViolation):
        eldar('Mother of the Eldar', 100.100)

# ....................{ TESTS ~ fail : param : hint        }....................
def test_nonpep_param_hint_invalid_fail() -> None:
    '''
    Test type-checking for a function with a parameter annotated with a type
    hint that is neither PEP-compliant *nor* PEP-noncompliant.
    '''

    # Import this decorator.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintNonpepException
    from beartype_test._util.pytroar import raises_uncached

    # Assert that type-checking a function with an integer parameter annotation
    # raises the expected exception.
    with raises_uncached(BeartypeDecorHintNonpepException):
        @beartype
        def nurgle(nurgling: str, great_unclean_one: 0x8BADF00D) -> str:
            return nurgling + str(great_unclean_one)

# ....................{ TESTS ~ fail : return              }....................
def test_nonpep_fail_return_call() -> None:
    '''
    Test type-checking for a function call unsuccessfully returning a value
    annotated with a PEP-noncompliant type hint.
    '''

    # Import this decorator.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintReturnViolation
    from pytest import raises

    # Annotated function to be type-checked.
    @beartype
    def necron(star_god: str, old_one: str) -> str:
        return 60e6

    # Call this function and assert the expected exception.
    with raises(BeartypeCallHintReturnViolation):
        necron("C'tan", 'Elder Thing')


def test_nonpep_fail_return_hint_nonpep() -> None:
    '''
    Test type-checking for a function with a return value unsuccessfully
    annotated with a type hint that is neither PEP-compliant *nor*
    PEP-noncompliant.
    '''

    # Import this decorator.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintNonpepException
    from beartype_test._util.pytroar import raises_uncached

    # Assert the expected exception from attempting to type-check a function
    # with a return annotation that is *NOT* a supported type.
    with raises_uncached(BeartypeDecorHintNonpepException):
        @beartype
        def tzeentch(disc: str, lord_of_change: str) -> 0xB16B00B5:
            return len(disc + lord_of_change)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **beartype import hookable coroutine submodule** (i.e., data module
containing *only* annotated asynchronous functions, mimicking real-world usage
of the :func:`beartype.claw.beartype_package` import hook from an external
caller).

See Also
--------
:mod:`beartype_test.a00_unit.data.claw.intraprocess.hookable_package.async.data_claw_func`
    Comparable beartype import hookable synchronous function submodule.
'''

# ....................{ TODO                               }....................
#FIXME: Also type-check asynchronous generators, please. *shrug*

# ....................{ IMPORTS                            }....................
from asyncio import (
    run as asyncio_run,
)
from beartype.typing import (
    Optional,
    Union,
)

# ....................{ COROUTINES                         }....................
async def _shone(not_a_sound_was_heard: Union[float, bytes]) -> (
    Optional[complex]):
    '''
    Arbitrary coroutine either returning the passed float first doubled and then
    coerced into a complex number with imaginary component ``1`` if this float
    is non-zero *or* raising a :exc:`.BeartypeCallHintParamViolation` exception
    otherwise (i.e., if this float is zero), exercising that beartype import
    hooks decorate global coroutines as expected.
    '''

    async def the_very_winds(dangers_grim_playmates: Optional[float]) -> (
        Union[complex, str]):
        '''
        Arbitrary coroutine closure either returning the passed float first
        doubled and then coerced into a complex number with imaginary component
        ``1`` if this float is non-:data:`None` *or* raising a
        :exc:`.BeartypeCallHintReturnViolation` exception otherwise (i.e., if
        this float is :data:`None`), exercising that beartype import hooks
        decorate coroutine closures as expected.
        '''

        # Return either...
        return (
            # If this float is "None", explicitly force this
            # @beartype-decorated closure to raise a
            # "BeartypeCallHintReturnViolation" exception;
            None
            if dangers_grim_playmates is None else
            # Else, this float is non-"None". In this case, return this float
            # doubled and then coerced into a complex number with imaginary
            # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
            not_a_sound_was_heard + dangers_grim_playmates + 1j
        )

    # Return either...
    return (
        # If this float is zero, explicitly force this @beartype-decorated
        # coroutine closure to raise a "BeartypeCallHintReturnViolation"
        # exception.
        await the_very_winds(None)
        if not_a_sound_was_heard == 0.0 else
        # Else, this integer is non-zero. In this case, return this integer
        # doubled and then coerced into a complex number with imaginary
        # component "1". Why? Just because. *THIS IS BEARTYPE.* Graaaah!
        await the_very_winds(not_a_sound_was_heard)
    )

# ....................{ MAIN                               }....................
async def _main() -> None:
    '''
    **Main coroutine runner** (i.e., high-level asynchronous function
    asynchronously exercising *all* lower-level coroutines defined above).

    This coroutine generator is intended to be (in order):

    #. Called synchronously.
    #. The coroutine implicitly created and returned by this coroutine generator
       passed to the standard :func:`asyncio.run` function, which then
       asynchronously runs the body of this coroutine as expected.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer coroutine-specific imports.
    from beartype import (
        beartype,
        BeartypeConf,
    )
    from beartype.roar import (
        BeartypeClawDecorWarning,
        BeartypeCallHintParamViolation,
        BeartypeCallHintReturnViolation,
        BeartypeDecorHintPep484Exception,
    )
    from beartype.typing import NoReturn
    from pytest import (
        raises,
        warns,
    )

    # ....................{ PASS                           }....................
    # Assert that calling this coroutine passed an arbitrary float returns the
    # expected complex number *WITHOUT* raising an exception.
    assert await _shone(
        len("Danger's grim playmates, on that precipice") + 0.0) == 84 + 1j

    # ....................{ FAIL                           }....................
    # Assert that calling this coroutine passed an invalid parameter raises the
    # expected exception.
    with raises(BeartypeCallHintParamViolation):
        await _shone(len('Slept, clasped in his embrace.—O, storm of death!'))

    # Assert that calling this function passed zero raises the expected
    # exception from the coroutine closure defined and called by this function.
    with raises(BeartypeCallHintReturnViolation):
        await _shone(
            len('Whose sightless speed.....') -       # <-- same string length *wat*
            len('divides this sullen night:') + 0.0)  # <-- same string length *lol*

    # ....................{ FAIL ~ decoration              }....................
    # Assert that attempting to define a coroutine whose type hints violate one
    # or more PEP standards at decoration time emits the expected non-fatal
    # warning.
    with warns(BeartypeClawDecorWarning):
        async def clasped_in_his_embrace(storm_of_death: NoReturn) -> None:
            '''
            Arbitrary coroutine accepting a parameter erroneously annotated with
            the :pep:`484`-compliant :obj:`typing.NoReturn` type hint *only*
            allowed on returns and thus raising a decoration-time exception,
            which the :mod:`beartype.claw` API then coerces into a non-fatal
            warning.
            '''

            pass

    # Assert that attempting to define a coroutine whose type hints violate one
    # or more PEP standards at decoration time emits the expected exception when
    # that function is configured by a non-default configuration instructing the
    # "beartype.claw" API to raise exceptions rather than emit non-fatal
    # warnings at decoration time. *PHEW!*
    with raises(BeartypeDecorHintPep484Exception):
        @beartype(conf=BeartypeConf(warning_cls_on_decorator_exception=None))
        async def whose_sightless_speed(
            divides_this_sullen_night: NoReturn) -> None:
            '''
            Arbitrary coroutine accepting a parameter erroneously annotated with
            the :pep:`484`-compliant :obj:`typing.NoReturn` type hint *only*
            allowed on returns and thus raising a decoration-time exception.
            '''

            pass


# Synchronously run the main coroutine runner defined above.
asyncio_run(_main())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **intraprocess import hook LangChain integration tests** (i.e.,
validating :mod:`beartype.claw` import hooks with respect to the third-party
LangChain API within the active Python process).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Isolate each integration test defined below to its own subprocess.
# Why? Module imports. Since each integration test defined below tends to
# reimport the same (or, at least, similar) modules as previously run unit tests
# defined below, module imports and thus unit tests *MUST* be isolated to their
# own subprocesses to ensure these tests may be run in any arbitrary order.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import pytest
from beartype_test._util.mark.pytskip import skip_unless_package

# ....................{ TESTS                              }....................
@skip_unless_package('langchain_core')
@pytest.mark.run_in_subprocess
def test_claw_langchain() -> None:
    '''
    Test :mod:`beartype.claw` import hooks against a LangChain-specific data
    subpackage in this test suite exercising these hooks against the third-party
    LangChain API known to publish decorator-hostile decorators.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.claw import beartype_package

    # ....................{ PASS                           }....................
    # Subject a LangChain-specific data package to a beartype import hook,
    # configured by the default beartype configuration.
    beartype_package('beartype_test.a90_func.data.claw.external.langchain')

    # Import all submodules of the package hooked above, exercising that these
    # submodules are subject to that import hook *AND* satisfy all integration
    # tests implemented as assertion statements in these submodules.
    from beartype_test.a90_func.data.claw.external.langchain.data_claw_langchain import (
        when_harboured_in)
    # print('[data_claw_langchain] when_harboured_in:')
    # print(when_harboured_in.__module__)
    # print(when_harboured_in.__class__.__name__)

    # Implicitly assert that the @beartype decorator avoids raising exceptions
    # (typically by internally reducing to a noop) when decorating uncallable
    # objects produced by LangChain-specific decorator-hostile decorators.
    when_harboured_in_beartyped = beartype(when_harboured_in)

    # Assert that the @beartype decorator actually did reduce to a noop.
    assert when_harboured_in is when_harboured_in_beartyped



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **TTY** (i.e., interactive terminal expected to be reasonably
POSIX-compliant, which even recent post-Windows 10 terminals now guarantee)
utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
import sys

# ....................{ TESTERS                            }....................
#FIXME: Unit test us up, please.
def is_stdout_terminal() -> bool:
    '''
    :data:`True` only if standard output is currently attached to a **TTY**
    (i.e., interactive terminal).

    If this tester returns :data:`True`, the TTY to which standard output is
    currently attached may be safely assumed to support **ANSI escape
    sequences** (i.e., POSIX-compliant colour codes). This assumption even holds
    under platforms that are otherwise *not* POSIX-compliant, including:

    * All popular terminals (including the stock Windows Terminal) and
      interactive development environments (IDEs) (including VSCode) bundled
      with Microsoft Windows, beginning at Windows 10.

    Caveats
    ----------
    **This tester is intentionally not memoized** (i.e., via the
    :func:`beartype._util.cache.utilcachecall.callable_cached` decorator), as
    external callers can and frequently do monkey-patch or otherwise modify the
    value of the global :attr:`sys.stdout` output stream.

    See Also
    ----------
    https://stackoverflow.com/questions/3818511/how-to-tell-if-python-script-is-being-run-in-a-terminal-or-via-gui
        StackOverflow thread strongly inspiring this implementation.
    '''
    # print(f'sys.stdout: {repr(sys.stdout)} [{type(sys.stdout)}]')
    # print(f'sys.stderr: {repr(sys.stderr)} [{type(sys.stderr)}]')

    # One-liners for great justice.
    #
    # Note that:
    # * Input and output streams are *NOT* guaranteed to define the isatty()
    #   method. For safety, we defensively test for the existence of that method
    #   before deferring to that method.
    # * All popular terminals under Windows >= 10 -- including terminals bundled
    #   out-of-the-box with Windows -- now support ANSII escape sequences. Since
    #   older Windows versions are compelling security risks and thus ignorable
    #   for contemporary purposes, Windows no longer needs to be excluded from
    #   ANSII-based colourization. All praise Satya Nadella. \o/
    return hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
    # return hasattr(sys.stdin, 'isatty') and sys.stdin.isatty()
    # return hasattr(sys.stderr, 'isatty') and sys.stderr.isatty()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) testers** (i.e.,
low-level callables testing and validating high-level
:class:`beartype.door.TypeHint` wrappers).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDoorException

# ....................{ VALIDATORS                         }....................
def die_unless_typehint(obj: object) -> None:
    '''
    Raise an exception unless the passed object is a **type hint wrapper**
    (i.e., :class:`TypeHint` instance).

    Parameters
    ----------
    obj : object
        Arbitrary object to be validated.

    Raises
    ------
    beartype.roar.BeartypeDoorException
        If this object is *not* a type hint wrapper.
    '''

    # Avoid circular import dependencies.
    from beartype.door._cls.doorsuper import TypeHint

    # If this object is *NOT* a type hint wrapper, raise an exception.
    if not isinstance(obj, TypeHint):
        raise BeartypeDoorException(
            f'{repr(obj)} not type hint wrapper '
            f'(i.e., "beartype.door.TypeHint" instance).'
        )
    # Else, this object is a type hint wrapper.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator class decoration** unit tests.

This submodule unit tests high-level functionality of the
:func:`beartype.beartype` decorator with respect to decorating **classes**
irrespective of lower-level type hinting concerns (e.g., PEP-compliance and
-noncompliance).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import (
    skip_if_python_version_greater_than_or_equal_to,
    skip_if_python_version_less_than,
)

# ....................{ TESTS ~ builtin                    }....................
def test_decor_type_descriptor_builtin() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on **C-based unbound builtin
    method descriptors** (i.e., methods decorated by builtin method decorators).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import NoReturn
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class OrDidASeaOfFire(object):
        '''
        Arbitrary class declaring a class method, static method, property
        getter, property setter, and property deleter -- all decorated by the
        :func:`beartype.beartype` decorator.
        '''

        # ....................{ CLASS VARS                 }....................
        so_much_of_life_and_joy_is_lost = 'all seems eternal now.'
        '''
        Arbitrary class variable.
        '''

        # ....................{ DUNDERS                    }....................
        def __init__(self) -> None:
            '''
            Arbitrary constructor, defined merely to set an instance variable
            subsequently manipulated by property methods defined below.
            '''

            self._taught_her_young_ruin = (
                'Ruin? Were these their toys? or did a sea')

        # ....................{ CLASS METHODS              }....................
        @beartype
        @classmethod
        def envelop_once_this_silent_snow(cls) -> str:
            '''
            Arbitrary class method decorated first by the builtin
            :class:`classmethod` descriptor and then by the
            :func:`beartype.beartype` decorator, exercising an edge case in the
            :func:`beartype.beartype` decorator.

            Note that reversing this order of decoration does *not* exercise an
            edge case in the :func:`beartype.beartype` decorator and is thus
            omitted from testing.
            '''

            return f'None can reply—{cls.so_much_of_life_and_joy_is_lost}'

        # ....................{ STATIC METHODS             }....................
        @beartype
        @staticmethod
        def the_wilderness_has_a_mysterious_tongue(
            of_man_flies_far_in_dread: str = '') -> str:
            '''
            Arbitrary static method decorated first by the builtin
            :class:`staticmethod` descriptor and then by the
            :func:`beartype.beartype` decorator, exercising an edge case in the
            :func:`beartype.beartype` decorator.

            Note that reversing this order of decoration does *not* exercise an
            edge case in the :func:`beartype.beartype` decorator and is thus
            omitted from testing.
            '''

            return f'Which teaches awful doubt,{of_man_flies_far_in_dread}'

        # ....................{ PROPERTIES                 }....................
        @beartype
        @property
        def where_the_old_earthquake_daemon(self) -> str:
            '''
            Arbitrary property getter method decorated first by the builtin
            :class:`staticmethod` descriptor and then by the
            :func:`beartype.beartype` decorator, exercising an edge case in the
            :func:`beartype.beartype` decorator.

            Note that reversing this order of decoration does *not* exercise an
            edge case in the :func:`beartype.beartype` decorator and is thus
            omitted from testing.
            '''

            return self._taught_her_young_ruin


        @beartype
        @where_the_old_earthquake_daemon.setter
        def where_the_old_earthquake_daemon(self, vanish: str) -> None:
            '''
            Arbitrary property setter method decorated first by the builtin
            :class:`staticmethod` descriptor and then by the
            :func:`beartype.beartype` decorator, exercising an edge case in the
            :func:`beartype.beartype` decorator.

            Note that reversing this order of decoration does *not* exercise an
            edge case in the :func:`beartype.beartype` decorator and is thus
            omitted from testing.
            '''

            self._taught_her_young_ruin = vanish


        @beartype
        @where_the_old_earthquake_daemon.deleter
        def where_the_old_earthquake_daemon(self) -> NoReturn:
            '''
            Arbitrary property deleter method decorated first by the builtin
            :class:`staticmethod` descriptor and then by the
            :func:`beartype.beartype` decorator, exercising an edge case in the
            :func:`beartype.beartype` decorator.

            Note that reversing this order of decoration does *not* exercise an
            edge case in the :func:`beartype.beartype` decorator and is thus
            omitted from testing.
            '''

            raise ValueError('And their place is not known.')

    # ....................{ LOCALS                         }....................
    # Instance of this class.
    the_race = OrDidASeaOfFire()

    # ....................{ PASS                           }....................
    # Assert this class method accessed on both this instance and this class
    # returns the expected value.
    assert (
        OrDidASeaOfFire.envelop_once_this_silent_snow() ==
        the_race.envelop_once_this_silent_snow() ==
        'None can reply—all seems eternal now.'
    )

    # Assert this class method docstring has been preserved as is.
    assert OrDidASeaOfFire.envelop_once_this_silent_snow.__doc__ is not None
    assert 'Arbitrary class method decorated first by the builtin' in (
       OrDidASeaOfFire.envelop_once_this_silent_snow.__doc__)

    # Assert this static method accessed on both this instance and this class
    # when passed a valid parameter returns the expected value.
    assert (
        OrDidASeaOfFire.the_wilderness_has_a_mysterious_tongue(' or faith so mild,') ==
        the_race.the_wilderness_has_a_mysterious_tongue(' or faith so mild,') ==
        'Which teaches awful doubt, or faith so mild,'
    )

    # Assert this static method docstring has been preserved as is.
    assert (
        OrDidASeaOfFire.the_wilderness_has_a_mysterious_tongue.__doc__ is not
        None
    )
    assert 'Arbitrary static method decorated first by the builtin' in (
        OrDidASeaOfFire.the_wilderness_has_a_mysterious_tongue.__doc__)

    # Assert this property getter method accessed on this instance returns the
    # expected value.
    assert the_race.where_the_old_earthquake_daemon == (
        'Ruin? Were these their toys? or did a sea')

    # Assert this property setter method accessed on this instance when passed a
    # valid parameter silently succeeds.
    the_race.where_the_old_earthquake_daemon = (
        "Vanish, like smoke before the tempest's stream,")

    # Assert this property getter method accessed on this instance returns this
    # valid parameter.
    assert the_race.where_the_old_earthquake_daemon == (
        "Vanish, like smoke before the tempest's stream,")

    # Assert this property method docstring has been preserved as is.
    assert OrDidASeaOfFire.where_the_old_earthquake_daemon.__doc__ is not None
    assert 'Arbitrary property getter method decorated first by the builtin' in (
        OrDidASeaOfFire.where_the_old_earthquake_daemon.__doc__)

    # ....................{ FAIL                           }....................
    # Assert this static method accessed on both this instance and this class
    # when passed an invalid parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        OrDidASeaOfFire.the_wilderness_has_a_mysterious_tongue(
            b'his work and dwelling')
    with raises(BeartypeCallHintParamViolation):
        the_race.the_wilderness_has_a_mysterious_tongue(
            b'The voiceless lightning in these solitudes')

    # Assert this property setter method accessed on this instance when passed
    # an invalid parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        the_race.where_the_old_earthquake_daemon = (
            b'And their place is not known. Below, vast caves')

    # Assert this property deleter method accessed on this instance raises the
    # expected exception.
    with raises(ValueError) as exception_info:
        del the_race.where_the_old_earthquake_daemon

    # Exception message raised by the body of that block.
    exception_message = str(exception_info.value)

    # Assert this exception message is the expected string.
    assert exception_message == 'And their place is not known.'


def test_decor_type_descriptor_builtin_called() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on **explicitly called C-based
    unbound builtin method descriptors** (i.e., builtin method decorators that
    are explicitly called as functions rather than implicitly called as
    decorators).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from types import GenericAlias

    # ....................{ CLASSES                        }....................
    @beartype
    class AndMeetLoneDeath(object):
        '''
        Arbitrary :func:`beartype.beartype`-decorated class declaring methods
        defined by explicitly calling C-based builtin method descriptors.
        '''

        __class_getitem__ = classmethod(GenericAlias)
        '''
        Augment this user-defined class into a type hint factory via the
        standard one-liner leveraged throughout both the standard library and
        third-party packages.

        Note that the :class:`classmethod` decorator explicitly supports C-based
        callable types. Ergo, this one-liner exercises that
        :func:`beartype.beartype` supports both this common idiom *and* this
        decorator behaviour without raising unexpected exceptions at decoration
        time.
        '''


# If the active Python interpreter targets either Python 3.9.x *OR* 3.10.x, then
# chaining the @classmethod decorator into the @property decorator is permitted;
# else, doing so is prohibited. To avoid non-deterministic behaviour under both
# older and newer Python versions, avoid those versions.
@skip_if_python_version_less_than('3.10.0')
@skip_if_python_version_greater_than_or_equal_to('3.11.0')
def test_decor_type_descriptor_builtin_chain() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on chaining multiple **C-based
    unbound builtin method descriptors** (i.e., methods decorated by builtin
    method decorators) together -- notably, the builtin :class:`.classmethod`
    decorator into the builtin :class:`.property` decorator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype

    # ....................{ CLASSES                        }....................
    class ThanGemsOrGold(object):
        '''
        Arbitrary class defining an arbitrary class property.
        '''

        @beartype
        @classmethod
        @property
        def the_varying_roof_of_heaven(cls) -> str:
            '''
            Arbitrary **class property** (i.e., method decorated by chaining the
            builtin :class:`.classmethod` decorator into the builtin
            :class:`.property` decorator).
            '''

            return 'And the green earth lost in his heart its claims'

    # ....................{ LOCALS                         }....................
    # Instance of this class.
    to_love_and_wonder = ThanGemsOrGold()

    # ....................{ PASS                           }....................
    # Assert this class property accessed on both this instance and this class
    # returns the expected value.
    assert (
        ThanGemsOrGold.the_varying_roof_of_heaven ==
        to_love_and_wonder.the_varying_roof_of_heaven ==
        'And the green earth lost in his heart its claims'
    )

# ....................{ TESTS ~ custom                     }....................
def test_decor_type_descriptor_custom() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on **custom pure-Python method
    descriptors** (i.e., methods decorated by pure-Python types satisfying the
    descriptor protocol by declaring at least the ``__get__()`` dunder method).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype.typing import Optional
    from beartype._data.kind.datakindiota import SENTINEL
    from collections.abc import Callable
    from functools import wraps
    from pytest import raises

    # ....................{ DESCRIPTORS                    }....................
    class flexible_descriptor(object):
        '''
        **Flexible class or bound method descriptor** (i.e., pure-Python object
        whose ``__get__()`` dunder method transparently supports access of the
        underlying pure-Python callable decorated by this descriptor as both a
        class method *and* bound method).
        '''

        def __init__(self, func: Callable) -> None:
            '''
            Initialize this method descriptor with the passed function.
            '''

            # Classify all passed parameters.
            # self._func = beartype(func)
            self._func = func


        def __get__(
            self,
            func_self: object = SENTINEL,
            func_cls: Optional[type] = SENTINEL,
        ) -> Callable:
            '''
            **Flexible class or bound method descriptor getter** (i.e., method
            decorating the low-level callable originally decorated by this
            descriptor by being passed to the :meth:`__init__` method with a
            higher-level wrapper, which this method then returns).
            '''

            # First parameter passed to the low-level callable decorated by this
            # descriptor, conditionally depending on whether this callable is
            # called as:
            # * A class method, in which case this parameter is "cls".
            # * A bound method, in which case this parameter is "self".
            self_or_cls = func_self if func_self is not None else func_cls

            # @beartype
            @wraps(self._func)
            def _func_flexible(*args, **kwargs):
                '''
                High-level closure wrapping the low-level callable decorated by
                this descriptor with:

                * :func:`beartype.beartype`-based type-checking.
                * Functionality enabling that callable to be flexibly called as
                  either a class *or* bound method.
                '''

                return self._func(self_or_cls, *args, **kwargs)

            # Return this closure.
            return _func_flexible

    # ....................{ CLASSES                        }....................
    @beartype
    class AndStillTheseTwo(object):
        '''
        Arbitrary :func:`beartype.beartype`-decorated class defining a flexible
        descriptor.
        '''

        _in_cathedral_cavern = 'Like natural sculpture in cathedral cavern;'
        '''
        Arbitrary class variable.
        '''


        def __init__(self) -> None:
            '''
            Initialize this object.
            '''

            # Arbitrary instance variable.
            self._the_frozen_god = 'The frozen God still couchant on the earth,'


        @flexible_descriptor
        @beartype
        def were_postured_motionless(
            self, like_natural_sculpture: str) -> str:
            '''
            Arbitrary flexible method safely callable as either a class *or*
            bound method, defined via the flexible descriptor defined above.

            Note that this method is *directly* decorated by the
            :func:`beartype.beartype` decorator. This is non-ideal. Ideally,
            the ``_func_flexible`` closure dynamically created and returned by
            the :class:`flexible_descriptor.__get__` dunder method would instead
            be decorated by the :func:`beartype.beartype` decorator. Sadly, the
            latter approach fails. Why? Because that closure is *not* explicitly
            passed the ``self`` parameter passed to this method. However,
            decorating that closure with :func:`beartype.beartype` would cause
            :func:`beartype.beartype` to (in order):

            #. Unwrap that closure to this decorated
               :meth:`were_postured_motionless` method.
            #. Parse the signature of this decorated method as if that were also
               the signature of that closure.
            #. Erroneously generate type-checking code expecting the *second*
               (rather than *first*) parameter to be this passed
               ``like_natural_sculpture`` parameter.
            '''

            # Return the concatenation of the passed string parameter *AND*...
            return like_natural_sculpture + (
                # If this flexible method is called as a class method, this
                # class variable;
                self._in_cathedral_cavern
                if self is AndStillTheseTwo else
                # Else, this flexible method is called as a bound method. In
                # this case, this bound variable.
                self._the_frozen_god
            )

    # ....................{ LOCALS                         }....................
    # Instance of the above class.
    still_couchant_on_the_earth = AndStillTheseTwo()

    # ....................{ PASS                           }....................
    # Assert that calling this flexible method as a class method when passed a
    # valid parameter returns the expected value.
    assert AndStillTheseTwo.were_postured_motionless(
        'And still these two were postured motionless,') == (
        'And still these two were postured motionless,'
        'Like natural sculpture in cathedral cavern;'
    )

    #
    # Assert that calling this flexible method as a bound method of this
    # instance when passed a valid parameter returns the expected value.
    assert still_couchant_on_the_earth.were_postured_motionless(
        'And the sad Goddess weeping at his feet:') == (
        'And the sad Goddess weeping at his feet:'
        'The frozen God still couchant on the earth,'
    )

    # ....................{ FAIL                           }....................
    # Assert that calling this flexible method as a class method when passed an
    # invalid parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        AndStillTheseTwo.were_postured_motionless(
            b'Until at length old Saturn lifted up')

    # Assert that calling this flexible method as a bound method of this
    # instance when passed an invalid parameter raises the expected exception.
    with raises(BeartypeCallHintParamViolation):
        still_couchant_on_the_earth.were_postured_motionless(
            b'His faded eyes, and saw his kingdom gone,')


def test_decor_type_descriptor_custom_pep487() -> None:
    '''
    Test the :func:`beartype.beartype` decorator on :pep:`487`-compliant
    **custom pure-Python method descriptors** (i.e., methods decorated by
    pure-Python types satisfying the :pep:`487`-compliant descriptor protocol by
    declaring at least the ``__set_name__()`` dunder method).

    This unit test exercises an inscrutable `edge case`_ induced by
    :func:`beartype.beartype`-decorated classes defining at least two kinds of
    seemingly unrelated methods that nonetheless *could* interact harmfully:

    * A method decorated by a :pep:`487`-compliant custom pure-Python method
      descriptor declaring at least the ``__set_name__()`` dunder method.
    * A method annotated by a stringified forward reference to a currently
      undefined type, which :func:`beartype.beartype` then internally coerces
      into a **forward reference proxy** (i.e.,
      :class:`beartype._check.forward.reference.fwdrefabc.BeartypeForwardRefABC`
      subclass) encapsulating that undefined type.

    For obscure reasons that are difficult to fully explain within the expected
    lifetime of the Universe, this interaction causes the ``__set_name__()``
    dunder method of the descriptor decorating the first method to be
    erroneously called multiple times with erroneous owners. Although seemingly
    innocuous, this issue is indicative of an explosive increase in the space
    and time complexity associated with forward reference proxies. Ergo, this
    issue is actually quite serious. It is imperative this *never* happen again.

    .. _edge case:
       https://github.com/beartype/beartype/issues/488#issuecomment-2650865299
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.typing import Optional
    from collections.abc import Callable

    # ....................{ DESCRIPTORS                    }....................
    class setname_descriptor(object):
        '''
        **Setname method descriptor** (i.e., pure-Python object whose
        :pep:`487`-compliant :meth:`__set_name__` dunder method raises an
        exception when called more than once as a regression check against this
        issue silently reverting).

        Attributes
        ----------
        _is_set_name_called : bool
            :data:`True` only if Python has already called the
            :pep:`487`-compliant :meth:`__set_name__` dunder method.
        '''

        def __init__(self, func: Callable) -> None:
            '''
            Initialize this method descriptor by doing... absolutely nothing.
            '''

            # Record that Python has yet to call __set_name__().
            self._is_set_name_called = False


        def __call__(self) -> None:
            '''
            Call this method descriptor by doing... absolutely nothing.
            '''

            # Doing nothing never felt so good, so right, so... clean.
            pass


        def __set_name__(self, owner: type, name: str) -> None:
            '''
            :pep:`487`-compliant callback called by Python itself with metadata
            describing the current decoration of an external method by this
            descriptor.

            Parameters
            ----------
            owner : type
                Parent class declaring that external method.
            name : str
                Unqualified basename of that external method.
            '''

            # If Python has already called this dunder method, this issue has
            # reverted. In this case, raise an exception.
            if self._is_set_name_called:
                raise ValueError(
                    'PEP 487 setname_descriptor.__set_name__() dunder method '
                    'already called.'
                )
            # Else, this is Python's first call of this dunder method.

            # Record that Python has already called this dunder method.
            self._is_set_name_called = True

    # ....................{ CLASSES                        }....................
    @beartype
    class AndAllThoseActs(object):
        '''
        Arbitrary :func:`beartype.beartype`-decorated class defining:

        * A method decorated by a :pep:`487`-compliant custom pure-Python method
          descriptor declaring at least the ``__set_name__()`` dunder method.
        * A method annotated by a stringified forward reference to a currently
          undefined type.

        Note that the mere act of decorating this class by
        :func:`beartype.beartype` suffices to fully exercise this issue.
        Instantiating this class below is incidental to that.
        '''

        @setname_descriptor
        def which_deity_supreme(self) -> None:
            '''
            Arbitrary method decorated by a :pep:`487`-compliant custom
            pure-Python method descriptor declaring at least the
            ``__set_name__()`` dunder method.
            '''

            # Sometimes, nothing is the best thing.
            pass


        def doth_ease_its(self) -> 'Optional[HeartOfLove]':
            '''
            Arbitrary method annotated by a stringified forward reference to a
            currently undefined type.
            '''

            # Nothing: still is the best thing after all these lines of code.
            pass


    class HeartOfLove(object):
        '''
        Arbitrary class referred to by a stringified forward reference
        annotating a method of the previously defined class.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Instance of the above class.
    i_am_gone = AndAllThoseActs()

    # ....................{ PASS                           }....................
    # Assert that calling this descriptor-decorated method behaves as expected.
    assert i_am_gone.which_deity_supreme() is None

    # Assert that calling this forward reference-annotated method behaves as
    # expected.
    assert i_am_gone.doth_ease_its() is None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`647`- or :pep:`742`-compliant **unit tests**.

This submodule unit tests both :pep:`647` and :pep:`742` support implemented in
the :func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.10.0')
def test_decor_pep647() -> None:
    '''
    Test :pep:`647` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.10 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintReturnViolation,
        BeartypeDecorHintPep647Exception,
    )
    from beartype.typing import (
        List,
        TypeGuard,
        Tuple,
    )
    from pytest import raises

    # ....................{ FUNCTIONS                      }....................
    @beartype
    def is_typeguard_slow(lst: List[object]) -> TypeGuard[List[str]]:
        '''
        :pep:`647`-compliant type guard returning :data:`True` only if the
        passed list contains only string items via the standard linear runtime
        type-check implemented manually.

        *leisurely barfs into adjacent bucket*
        '''

        return all(isinstance(item, str) for item in lst)


    @beartype
    def is_typeguard_bad(tup: Tuple[object, ...]) -> TypeGuard[Tuple[int, ...]]:
        '''
        :pep:`647`-compliant type guard erroneously (presumably, accidentally)
        violating :pep:`647` by returning a non-boolean object.
        '''

        return "Lull'd by the coil of his crystalline streams,"

    # ....................{ PASS                           }....................
    # Assert that the above function passed lists both satisfying and violating
    # this type guard returns the expected booleans.
    assert is_typeguard_slow(
        ['Thou', 'who', 'didst', 'waken', 'from', 'his', 'summer', 'dreams']
    ) is True
    assert is_typeguard_slow(
        ['Thou', 'who', 'didst', 'waken', 'from', 'his', 'summer', b'dreams']
    ) is False

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating a
    # callable accepting one or more parameters erroneously annotated by a PEP
    # 647-compliant type guard.
    with raises(BeartypeDecorHintPep647Exception):
        @beartype
        def the_blue_mediterranean(where_he_lay: TypeGuard[str]) -> bool:
            return where_he_lay

    # Assert that @beartype raises the expected violation when calling a
    # callable erroneously violating a PEP 647-compliant type guard.
    with raises(BeartypeCallHintReturnViolation):
        is_typeguard_bad(('Beside a', 'pumice isle', 'in', "Baiae's bay,",))


@skip_if_python_version_less_than('3.13.0')
def test_decor_pep742() -> None:
    '''
    Test :pep:`742` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.13 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import (
        BeartypeCallHintReturnViolation,
        BeartypeDecorHintPep742Exception,
    )
    from beartype.typing import (
        List,
        TypeIs,
        Tuple,
    )
    from pytest import raises

    # ....................{ FUNCTIONS                      }....................
    @beartype
    def is_typeguard_slow(lst: List[object]) -> TypeIs[List[str]]:
        '''
        :pep:`742`-compliant type guard returning :data:`True` only if the
        passed list contains only string items via the standard linear runtime
        type-check implemented manually.

        *leisurely barfs into adjacent bucket*
        '''

        return all(isinstance(item, str) for item in lst)


    @beartype
    def is_typeguard_bad(tup: Tuple[object, ...]) -> TypeIs[Tuple[int, ...]]:
        '''
        :pep:`742`-compliant type guard erroneously (presumably, accidentally)
        violating :pep:`742` by returning a non-boolean object.
        '''

        return 'He hath prepared, prowling around the world;'

    # ....................{ PASS                           }....................
    # Assert that the above function passed lists both satisfying and violating
    # this type guard returns the expected booleans.
    assert is_typeguard_slow(
        ['Glutted', 'with', 'which', 'thou', 'mayst', 'repose,', 'and', 'men']
    ) is True
    assert is_typeguard_slow(
        ['Glutted', 'with', 'which', 'thou', 'mayst', 'repose,', 'and', b'men']
    ) is False

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating a
    # callable accepting one or more parameters erroneously annotated by a PEP
    # 742-compliant type guard.
    with raises(BeartypeDecorHintPep742Exception):
        @beartype
        def he_hath_prepared(prowling_around_the_world: TypeIs[str]) -> bool:
            return prowling_around_the_world

    # Assert that @beartype raises the expected violation when calling a
    # callable erroneously violating a PEP 742-compliant type guard.
    with raises(BeartypeCallHintReturnViolation):
        is_typeguard_bad((
            'Go', 'to', 'their', 'graves', 'like', 'flowers',
            'or', 'creeping', 'worms,'
        ))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`604`-compliant **union type hint utility**
unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484604` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep604() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484604.is_hint_pep604` tester.
    '''

    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep484604 import is_hint_pep604

    # Assert this tester accepts a PEP 604-compliant union.
    assert is_hint_pep604(int | str | None) is True

    # Assert this tester rejects an arbitrary PEP 604-noncompliant object.
    is_hint_pep604('Meet in the vale, and one majestic River,')

# ....................{ TESTS ~ factory                    }....................
def test_make_hint_pep484604_union() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep484604.make_hint_pep484604_union`
    factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep604Exception
    from beartype.typing import (
        Union,
    )
    from beartype._util.hint.pep.proposal.pep484604 import (
        make_hint_pep484604_union)
    from pytest import raises
    from typing import Annotated

    # ....................{ PASS                           }....................
    # Assert that this factory when passed a 1-tuple containing an arbitrary
    # child type hint simply returns that hint as is.
    assert make_hint_pep484604_union((int,)) is int

    # Assert that this factory when passed a tuple containing two or more PEP
    # 604-compliant child type hints returns a PEP 604-compliant new-style union
    # of these hints.
    assert make_hint_pep484604_union((int, list[str], None)) == (
        int | list[str] | None)

    # Assert that this factory when passed a tuple containing two or more
    # child type hints at least one of which is PEP 593-compliant (and thus
    # *NOT* PEP 604-compliant) returns a PEP 484-compliant old-style union of
    # these hints.
    assert make_hint_pep484604_union(
        (bool, list[int], None, Annotated[bytes, None])) == Union[
         bool, list[int], None, Annotated[bytes, None]]

    # ....................{ FAIL                           }....................
    # Assert that this factory raises the expected exception when passed the
    # empty tuple.
    with raises(BeartypeDecorHintPep604Exception):
        make_hint_pep484604_union(())



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-noncompliant type hints test data.**

This submodule predefines low-level global constants whose values are
PEP-noncompliant type hints, exercising known edge cases on behalf of
higher-level unit test submodules.
'''

# ....................{ IMPORTS                            }....................
from pytest import fixture

# ....................{ FIXTURES                           }....................
@fixture(scope='session')
def hints_nonpep_meta() -> 'Tuple[HintNonpepMetadata]':
    '''
    Session-scoped fixture yielding a tuple of **PEP-noncompliant type hint
    metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintNonpepMetadata`
    instances, each describing a sample PEP-noncompliant type hint exercising an
    edge case in the :mod:`beartype` codebase).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintNonpepMetadata)
    from beartype_test._util.kind.pytkindmake import make_container_from_funcs

    # ..................{ LIST                               }..................
    _hints_nonpep_meta = make_container_from_funcs((
        # PEP-compliant type hints.
        'beartype_test.a00_unit.data.hint.nonpep.beartype._data_nonpepbeartype.hints_nonpepbeartype_meta',
        'beartype_test.a00_unit.data.hint.nonpep.proposal._data_nonpep484.hints_nonpep484_meta',
        'beartype_test.a00_unit.data.hint.nonpep.proposal._data_nonpep695.hints_nonpep695_meta',
    ))

    # ..................{ YIELD                              }..................
    # Assert this list contains *ONLY* instances of the expected dataclass.
    assert (
        isinstance(hint_nonpep_meta, HintNonpepMetadata)
        for hint_nonpep_meta in _hints_nonpep_meta
    ), (f'{repr(_hints_nonpep_meta)} not iterable of '
        f'"HintNonpepMetadata" instances.')

    # Yield a tuple coerced from this list.
    yield tuple(_hints_nonpep_meta)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`585`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep585Exception
from beartype.typing import (
    Dict,
    TypeVar,
)
from beartype._cave._cavefast import HintGenericSubscriptedType
from beartype._data.typing.datatypingport import (
    Hint,
    TupleHints,
)
from beartype._data.typing.datatyping import (
    TupleTypeVars,
    TypeException,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.kind.maplike.utilmapset import update_mapping_keys
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11

# ....................{ HINTS                              }....................
HINT_PEP585_TUPLE_EMPTY = tuple[()]
'''
:pep:`585`-compliant empty fixed-length tuple type hint.
'''

# ....................{ RAISERS                            }....................
def die_unless_hint_pep585_generic(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep585Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception unless the passed object is a :pep:`585`-compliant
    **generic** (i.e., either a type originally subclassing at least one
    subscripted :pep:`585`-compliant pseudo-superclass *or* an object
    subscripted by one or more child type hints originating from such a type).

    This raiser raises an exception unless this object is either a subscripted
    or unsubscripted :pep:`585`-compliant generic.

    Parameters
    ----------
    hint : Hint
        Object to be validated.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep585Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Raises
    ------
    exception_cls
        If this object is *not* a :pep:`585`-compliant generic.
    '''

    # If this object is *NOT* a PEP 585-compliant generic, raise an exception.
    if not is_hint_pep585_generic(hint):
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not PEP 585 generic.')
    # Else, this object is a PEP 585-compliant generic.

# ....................{ TESTERS                            }....................
def is_hint_pep585_builtin_subbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`585`-compliant
    **subscripted builtin type hint** (i.e., C-based type hint instantiated by
    subscripting either a concrete builtin container class like :class:`list` or
    :class:`tuple` *or* an abstract base class (ABC) declared by the
    :mod:`collections.abc` submodule like :class:`collections.abc.Iterable` or
    :class:`collections.abc.Sequence`).

    This tester additionally returns :data:`True` for third-party type hints
    whose types subclass the :class:`types.GenericAlias` superclass, including:

    * ``numpy.typing.NDArray[...]`` type hints.

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This tester returns** :data:`False` for :pep:`585`-compliant generics,
    which fail to satisfy the same API as all other :pep:`585`-compliant type
    hints. Why? Because :pep:`560`-type erasure erases the low-level superclass
    detected by this tester on :pep:`585`-compliant generics immediately after
    those generics are declared, preventing their subsequent detection as
    :pep:`585`-compliant. Instead, :pep:`585`-compliant generics are only
    detectable by calling either:

    * The high-level PEP-agnostic
      :func:`beartype._util.hint.pep.utilpeptest.is_hint_pep484585_generic`
      tester.
    * The low-level :pep:`585`-specific :func:`.is_hint_pep585_generic` tester.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`585`-compliant type hint.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585gentest import (
        is_hint_pep484585_generic)

    # Return true only if this hint...
    return (
        # Is either a PEP 484- or -585-compliant subscripted generic or
        # PEP 585-compliant builtin *AND*...
        isinstance(hint, HintGenericSubscriptedType) and
        # Is *NOT* a PEP 484- or -585-compliant subscripted generic.
        not is_hint_pep484585_generic(hint)  # pyright: ignore
    )


def is_hint_pep585_generic(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`585`-compliant **generic**
    (i.e., either a type originally subclassing at least one subscripted
    :pep:`585`-compliant pseudo-superclass *or* an object subscripted by one or
    more child type hints originating from such a type).

    This tester returns :data:`True` if this object is either a subscripted or
    unsubscripted :pep:`585`-compliant generic.

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`585`-compliant generic.
    '''

    # Return true only if object is either...
    return (
        # A PEP 585-compliant unsubscripted generic *OR*...
        is_hint_pep585_generic_unsubbed(hint) or
        # A PEP 585-compliant subscripted generic.
        is_hint_pep585_generic_subbed(hint)
    )


#FIXME: Unit test us up, please.
def is_hint_pep585_generic_subbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`585`-compliant
    **subscripted generic** (i.e., object subscripted by one or more child type
    hints originating from a type originally subclassing at least one
    subscripted :pep:`585`-compliant pseudo-superclass).

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`585`-compliant subscripted
        generic.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_origin_or_none)

    # Arbitrary object originating this hint if any *OR* "None" otherwise.
    hint_origin = get_hint_pep_origin_or_none(hint)

    # Return true only if...
    return (
        # An object originates this hint *AND*...
        hint_origin is not None and
        # This origin object is an unsubscripted generic type, which would then
        # imply this hint to be a subscripted generic. If this strikes you as
        # insane, you're not alone.
        is_hint_pep585_generic_unsubbed(hint_origin)
    )


#FIXME: Unit test us up, please.
@callable_cached
def is_hint_pep585_generic_unsubbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is a :pep:`585`-compliant
    **unsubscripted generic** (i.e., type originally subclassing at least one
    subscripted :pep:`585`-compliant pseudo-superclass).

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`585`-compliant unsubscripted
        generic.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args
    from beartype._util.hint.pep.proposal.pep560 import (
        is_hint_pep560,
        iter_hint_pep560_bases_unerased,
    )

    # If it is *NOT* the case that...
    if not (
        # This hint is a type *AND*...
        isinstance(hint, type) and
        # This type is PEP 560-compliant and thus subclasses one or more
        # pseudo-superclasses *AND*...
        is_hint_pep560(hint) and
        # Either...
        (
            # The active Python interpreter targets Python >= 3.11 *OR*...
            IS_PYTHON_AT_LEAST_3_11 or
            # The active Python interpreter targets Python <= 3.10. In this
            # case, a simple subclass test does *NOT* suffice to detect a PEP
            # 585-compliant unsubscripted generic. Why? Because Python <=
            # 3.10 implements PEP 585-compliant subscripted generics as types!
            # But PEP 585-compliant unsubscripted generics are also types! Lo:
            #     $ python3.10
            #     >>> class MuhGeneric(list): pass
            #     >>> isinstance(MuhGeneric, type)
            #     True  # <-- good
            #     >>> isinstance(MuhGeneric[str], type)
            #     True  # <-- *BAD*
            #
            #     $ python3.11
            #     >>> class MuhGeneric(list): pass
            #     >>> isinstance(MuhGeneric, type)
            #     True  # <-- good
            #     >>> isinstance(MuhGeneric[str], type)
            #     False  # <-- good
            #
            # Disambiguating this edge case requires also detecting whether
            # this PEP 484-compliant generic is subscripted by one or more
            # child hints.
            #
            # This PEP 484-compliant generic is unsubscripted.
            not get_hint_pep_args(hint)
        )
    # Then this hint *CANNOT* be a PEP 585-compliant unsubscripted generic. In
    # this case, return false immediately.
    ):
        return False
    # Else, this hint is a PEP 560-compliant type subclassing one or more
    # pseudo-superclasses. Since this type *COULD* be a PEP 585-compliant
    # unsubscripted generic, continue testing.

    #FIXME: [SPEED] Optimize into a "while" loop. *sigh*
    # For each transitive pseudo-superclass of this PEP 560-compliant hint...
    #
    # Unsurprisingly, PEP 585-compliant generics have absolutely *NO*
    # commonality with PEP 484-compliant generics. While the latter are
    # trivially detectable as subclassing "typing.Generic" after type erasure,
    # the former are *NOT*. The only means of deterministically deciding whether
    # or not a hint is a PEP 585-compliant generic is if:
    # * That class defines both the __class_getitem__() dunder method *AND* the
    #   "__orig_bases__" instance variable. Note that this condition in and of
    #   itself is insufficient to decide PEP 585-compliance as a generic. Why?
    #   Because these dunder attributes have been standardized under various
    #   PEPs and may thus be implemented by *ANY* arbitrary classes.
    # * The "__orig_bases__" instance variable is a non-empty tuple.
    # * One or more objects listed in that tuple are PEP 585-compliant C-based
    #   subscripted generics (e.g., "list[str]").
    #
    # Note we could technically also test that this hint defines the
    # __class_getitem__() dunder method. Since that test would *NOT* suffice to
    # ensure that this hint is a PEP 585-compliant generic, however, there
    # exists little benefit to doing so.
    for hint_base in iter_hint_pep560_bases_unerased(hint):
        # If this transitive pseudo-superclass is itself a PEP 585-compliant
        # subscripted generic (e.g., "list[str]"), the passed hint transitively
        # subclasses a PEP 585-compliant generic. By transitivity, this hint
        # *MUST* be a PEP 585-compliant generic as well. Return true!
        if is_hint_pep585_builtin_subbed(hint_base):
            return True
        # Else, this pseudo-superclass is *NOT* a PEP 585-compliant subscripted
        # generic. In this case, continue to the next pseudo-superclass.

    # Since *NO* such pseudo-superclasses are PEP 585-compliant subscripted
    # generics, this hint is *NOT* a PEP 585-compliant generic. In this case,
    # return false.
    return False

# ....................{ GETTERS                            }....................
def get_hint_pep585_generic_bases_unerased(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep585Exception,
    exception_prefix: str = '',
) -> TupleHints:
    '''
    Tuple of all unerased :pep:`585`-compliant **pseudo-superclasses** (i.e.,
    :mod:`typing` objects originally listed as superclasses prior to their
    implicit type erasure under :pep:`560`) of the passed :pep:`585`-compliant
    **generic** (i.e., class subclassing at least one non-class
    :pep:`585`-compliant object).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep585Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    Tuple[Hint, ...]
        Tuple of the one or more unerased pseudo-superclasses of this
        :pep:`585`-compliant generic.

    Raises
    ------
    exception_cls
        If this hint is *not* a :pep:`585`-compliant generic.

    See Also
    --------
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_bases_unerased`
        Further details.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_type_or_none)

    # If this hint is *NOT* a class, reduce this hint to the object originating
    # this hint if any. See the is_hint_pep484_generic() tester for details.
    hint = get_hint_pep484585_generic_type_or_none(hint)  # type: ignore[assignment]

    # If this hint is *NOT* a PEP 585-compliant generic, raise an exception.
    die_unless_hint_pep585_generic(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is a PEP 585-compliant generic.

    # Return the tuple of all unerased pseudo-superclasses of this generic.
    # While the "__orig_bases__" dunder instance variable is *NOT* guaranteed
    # to exist for PEP 484-compliant generic types, this variable is guaranteed
    # to exist for PEP 585-compliant generic types. Thanks for small favours.
    return hint.__orig_bases__  # pyright: ignore


@callable_cached
def get_hint_pep585_generic_typeargs_packed(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep585Exception,
    exception_prefix: str = '',
) -> TupleTypeVars:
    '''
    Tuple of all **unique type variables** (i.e., non-duplicated
    :class:`TypeVar` objects) originally parametrizing the passed
    :pep:`585`-compliant unsubscripted generic if this generic was parametrized
    by one or more type variables *or* the empty tuple otherwise (i.e., if this
    generic is unparametrized).

    This getter is memoized for efficiency.

    Motivation
    ----------
    This getter mimics the behaviour of the ``__parameters__`` dunder attributes
    for :pep:`484`-compliant generics, whose values similarly collect the tuples
    of all unique type variables originally parametrizing those generics. Sadly,
    the current implementation of :pep:`585` under at least Python 3.9—3.13 is
    fundamentally broken with respect to parametrized generics. While
    :pep:`484`-compliant generics properly propagate type variables from
    pseudo-superclasses to subclasses, :pep:`585` fails to do so. This getter
    "fills in the gaps" by recovering these type variables from parametrized
    :pep:`585`-compliant generics by iteratively constructing a new tuple from
    the type variables parametrizing all pseudo-superclasses of this generic.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep585Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    Tuple[TypeVar, ...]
        Either:

        * If this :pep:`585`-compliant generic is:

          * Subscripted, the empty tuple. This mirrors the behaviour of the
            ``__parameters__`` dunder attribute defined on :pep:`484`-compliant
            subscripted generics, which is *always* set to the empty tuple.
          * Unsubscripted, the tuple of all unique type variables parametrizing
            this unsubscripted generic

        * Else, the empty tuple.

    Raises
    ------
    BeartypeDecorHintPep585Exception
        If this hint is *not* a :pep:`585`-compliant generic.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_typeargs_packed

    # If this hint is *NOT* a PEP 585-compliant generic, raise an exception.
    die_unless_hint_pep585_generic(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is a PEP 585-compliant generic.

    # Tuple of all type variables parametrizing this generic if this is a PEP
    # 585-compliant subscripted generic *OR* "None" otherwise (i.e., is a PEP
    # 585-compliant unsubscripted generic). For known reasons, the
    # "__parameters__" dunder attribute is defined correctly for PEP
    # 585-compliant subscripted (but *NOT* unsubscripted) generics. *shrug*
    hint_typevars = getattr(hint, '__parameters__', None)

    # If this tuple is defined, return this tuple as is.
    if hint_typevars is not None:
        return hint_typevars
    # Else, this tuple is undefined. In this case, synthetically reconstruct
    # this tuple for this PEP 585-compliant unsubscripted generic.

    # Tuple of all pseudo-superclasses of this unsubscripted generic.
    hint_bases = get_hint_pep585_generic_bases_unerased(hint)

    # Dictionary mapping from all type variables parametrizing these
    # pseudo-superclasses to the "None" singleton, thus preserving the ordering
    # of these type variables while yet discarding duplicate type variables
    # parametrizing multiple pseudo-superclasses.
    #
    # Note that:
    # * A dictionary rather than set is intentionally leveraged. Why? Ordering.
    #   Order of type variables is *EXTREMELY* significant (e.g., when mapping
    #   type variables to child hints in a type variable lookup table). Whereas
    #   dictionaries preserve insertion order as of Python >= 3.7, sets
    #   currently do *NOT*. Since this algorithm *ONLY* employs a dictionary to
    #   preserve insertion order, the values of this dictionary are irrelevant
    #   and thus unconditionally set to the "None" singleton. I sigh so hard.
    # * The following inefficient iteration *CANNOT* be trivially reduced to a
    #   dictionary comprehension, as each get_hint_pep_typeargs_packed() call returns a
    #   tuple of type variables rather than a single type variable to be added
    #   to this dictionary.
    hint_typevars_to_none: Dict[TypeVar, None] = dict()

    # For each such pseudo-superclass...
    for hint_base in hint_bases:
        # Tuple of the zero or more type variables parametrizing this
        # pseudo-superclass.
        hint_base_typevars = get_hint_pep_typeargs_packed(hint_base)
        # print(f'hint_base_typevars: {hint_base} [{get_hint_pep_typeargs_packed(hint_base)}]')

        # Efficiently add these type variables as new keys of this dictionary.
        update_mapping_keys(hint_typevars_to_none, hint_base_typevars)

    # Return a tuple coerced from the keys of this dictionary.
    return tuple(hint_typevars_to_none.keys())
